#!/usr/bin/perl -w

use strict;

use IO::Select;
use IO::Socket;
use Getopt::Long;
use POSIX qw/floor/;

use Astro::Time;
use Astro::Coord;
use Astro::Vex;
use RtFC;

sub Initialize ();
sub Shutdown ();
sub child_reaper;
sub process_obs_command ($$$$);
sub process_obsdata_command ($$);
sub process_control_command ($);
sub setup_correlation ($);
sub correlate ($$$$$$@);
sub launch_corr ($$$$);
sub plotfringe($$$$$$\@\@@);
sub plotit ($$$$$$$$$@);

$Astro::Time::StrZero = 2;


# Global variables, yuck but they are quickly copied into a structure

my %offsets = ();
my %clockrates = ();
my $corr_lags = 128;
my $pack = 0;

my $control_host = defined($ENV{RTFC_CONTROL}) ?
                                         $ENV{RTFC_CONTROL}: 'localhost';

my $corr_host = defined($ENV{RTFC_CORRINTERFACE}) ?
                                         $ENV{RTFC_CORRINTERFACE}: 'localhost';
my $expand = 1;
my $nocorr = 0;
my $delayfile = undef;

GetOptions('control=s'=>\$control_host,'correlator=s'=>\$corr_host,
	   'expand!'=>\$expand, 'nocorr'=>\$nocorr, 
	   'delay=s'=>\$delayfile);

if (!@ARGV && ! $nocorr) {
  die "Usage: corr.pl <vexfile> [<vexfile> ...]\n";
}

if (!defined $delayfile) {
  die "Must specify delay file" if (!$nocorr);
  $delayfile = 'none';
} else {
  die "No such file $delayfile\n" if (! -e $delayfile);
  die "$delayfile is not a plain file\n" if (! -f $delayfile);
}

# Open the vexfile. This really should be a changeable value from the UI.

my @vex;
foreach (@ARGV) {
  push @vex, new Astro::Vex($_);

  if ((!defined $vex[$#vex])) {
    die "Error reading $_\n";
  }
}


$SIG{INT} = \&Shutdown;
$SIG{CHLD} = \&child_reaper;

my ($sock_obs, $sock_control) = Initialize();

# Loops for ever, waiting for data on the various sockets. We might
# also want to run periodic commands

my $sel = new IO::Select($sock_obs, $sock_control);

my @ready;

# Lists of observatory OBJECTS which are currently connected
my @Obs = ();
my @ObsData = ();

# Hash of outstanding (ie still downloading) correlation objects
my %Corrs = ();

# Get the timeranges for all the vex files
my @vextimes = ();

foreach (@vex) {
  my @scans = $_->sched();

  my $start = $scans[0]->start;
  my $finish = $start+$scans[0]->maxscan/(60*60*24);

  foreach (@scans) {
    if ($_->start < $start) {
      $start = $_->start;
    }
    if ($_->start + $_->maxscan/(60*60*24) > $finish) {
      $finish = $_->start + $_->maxscan/(60*60*24);
    }
  }

  push @vextimes, [$start, $finish];

  my ($day, $month, $year, $ut) = mjd2cal($start);
  my $start_str = sprintf("%02d/%02d/%04d %s", $day, $month, $year,
			  turn2str($ut, 'H', 0));
  ($day, $month, $year, $ut) = mjd2cal($finish);
  my $stop_ut = turn2str($ut, 'H', 0);

  printf("%s: $start_str  --  $stop_ut\n", $_->exper->exper_name);

}

# Set the xterm titlebar
print "\033]0;RtFC corr.pl\007\n";
$| = 1;  # Flush stdout

my $slept = 0;
while (1) {
  @ready = $sel->can_read(600);
  if (!@ready) {
    print '.';
    $slept = 1;
    next;
  }
  if ($slept) {
    print "\n";
    $slept = 0;
  }

  foreach my $fh (@ready) {
    my $handled = 0;
    if (defined $sock_obs && $fh == $sock_obs) {
      # Observatory connection request
      $handled = 1;
      
      print "Got a new observatory socket connection...\n";
      # Create a new socket
      my $newobs = $sock_obs->accept();
      push @Obs, new RtFC::ObsLink($newobs);
      $sel->add($newobs);
    } elsif ($fh == $sock_control) {
      # Control process message
      $handled = 1;
      
      my $command = recv_command($sock_control,$sel,'control');

      if (ref $command) { # Must be a command
	process_control_command($command);
      } else {
	next if (! defined $command); # There was a read error
	if ($command==-1) { # Socket was closed remotely
	  warn "Control socket gone. Shutting down...\n";
	  Shutdown();
	}
      }
    } else {
      # Must have data from one of the client obs sockets...
      
      my @killedsock = ();
      my $i=0;
      foreach my $obs (@Obs) {
	my $sock = $obs->socket;
	if ($fh==$sock) { # Yep, this is a obs socket with some data
	  $handled = 1;
	  
	  my $command = recv_command($sock,$sel,'observatory',$i,
				       \@killedsock);
	  
	  if (ref $command) { # Must be a command
	    process_obs_command($obs, $command, $i, \@killedsock);
	  } else {
	    # We can ignore these errors at the moment
	  }
	}
	$i++;
      }
      # Remove any obs sockets which were closed
      foreach (@killedsock) {
	splice @Obs, $_, 1; # Remove the i'th element
      }
      
      @killedsock = ();
      $i=0;
      foreach my $obs (@ObsData) { # Loop through the child connections
	my $sock = $obs->socket;
	if ($fh==$sock) { # Yep, this is a obs socket with some data
	  $handled = 1;
	  
	  my $command = recv_command($sock,$sel,'observatory data',$i,
				     \@killedsock);
	  
	  if (ref $command) { # Must be a command
	    process_obsdata_command($obs, $command);
	  } else {
	    # We can ignore these errors at the moment
	  }
	}
	$i++;
      }
      # Remove any parent sockets which were closed
      foreach (@killedsock) {
	splice @ObsData, $_, 1; # Remove the i'th element
      }
    }
    if (!$handled) {
      warn "Did not handle message from $fh\n";
    }
  }
}

sub process_obs_command ($$$$) {
  my ($obs, $command, $i, $killedsock) = @_;

  if ($command->type==ANTENNA_NAME_COMMAND) {
    printf "Connection is from %s\n", $command->antenna_name;
    $obs->name($command->antenna_name);

  } elsif ($command->type==ANTENNA_ID_COMMAND) {
    $obs->code($command->antenna_id);

  } elsif ($command->type eq REDEFINE_COMMAND) {
    if ($command->redef==OBSDATA) {
      # This is the secondary pipe for actually dumping the data
      # First, remove from @Obs list
      unshift @{$killedsock}, $i; # Store largest index first
      my $obsdata = create_data_process($obs);
      if (defined $obsdata) {
	push @ObsData, $obsdata;
	$sel->add($obsdata->socket);
      }
    } else {
      die "Cannot redefine connection to $command->redef at ";
    }
  } else {
    printf STDERR "Unknown command (%d) received\n", $command->type;
  }
}

sub process_obsdata_command ($$) {
  my ($obs, $command) = @_;

  if ($command->type==DATA_RECEIVED_COMMAND) {
    # Got some data. See what other data has arrived so we can correlate it

    my $code = $command->code;
    my $mjd = $command->mjd; # Should check wrt $corr->mjd
    my $dir = $command->dir;
    my $file = $command->file;
    my $mode = $command->mode;
    my $id = $obs->code;

    printf "Try and correlate data from %s (%s)\n", $id, "$dir/$file";
    if (exists $Corrs{$code}) {
      # Check this data is expected and not already received
      my $corr = $Corrs{$code};
      my $obs = $corr->obs;
      my @expected_obs = keys(%{$obs});

      # First check we expect data from this telescope
      my $found = 0;
      foreach (@expected_obs) {
	if ($_ eq $id) {
	  $found=1;
	  last;
	}
      }
      if ($found) {
	$obs->{$id} = "$dir/$file"; # Save this for later

	correlate($mjd, $corr, $obs, [$id], $dir, $mode, @expected_obs);

      } else {
	warn "Not expecting data from $id!! (@expected_obs)\n";
      }

    } else {
      warn "Received data from unknown request!!\n";
    }
  } else {
    printf STDERR "Unknown command (%d) received\n", $command->type;
  }
}

sub process_control_command ($) {
  my ($command) = @_;

  if ($command->type==ANTENNA_NAME_COMMAND) {
    printf "Connection is from %s\n", $command->antenna_name;
  } elsif ($command->type==ERROR_COMMAND) {
    if ($command->error==DUPLICATE_CONNECTION) {
      # Cannot connect as there is another connection already
      warn "Correlator connection already present. Aborting...\n";
      Shutdown();
    } else {
      printf STDERR "Unknown error code: %d\n", $command->error;
    }

  } elsif ($command->type==GRAB_COMMAND) {
    setup_correlation($command);

  } elsif ($command->type==CORRSETUP_COMMAND) {
    $corr_lags = $command->nlags;
    $pack = $command->pack;
    %offsets = ();
    foreach (@{$command->offsets}) {
      $offsets{lc($_->[0])} = $_->[1];
    }
    %clockrates = ();
    foreach (@{$command->clockrates}) {
      $clockrates{lc($_->[0])} = $_->[1];
    }
  } else {
    printf STDERR "Unknown command (%d) received\n", $command->type;
  }
}

sub create_data_process ($) {
  my $obs = shift;

  # We want to fork and create a local loop polling on the data pipe

  my $child = new IO::Socket;
  my $parent = new IO::Socket;

  # Create a socket connection between the parent and child
  socketpair($child, $parent, AF_UNIX, SOCK_STREAM, PF_UNSPEC)
    or  die "Could not create socketpair: $! at ";

  my $pid;

  if ($pid = fork) { # Parent fork
    $sel->remove($obs->socket);
    $obs->socket($child);
    $obs->pid($pid);
    return $obs;

  } else {
    # Child fork
    die "cannot fork: $! at " unless defined $pid;
    #$child->close;
    my $childmsg = sprintf 'Child(%s):', $obs->code;

    my ($buf);

    # Set up own own private loop accepting data from the observatory
    my @ready;
    my $data_pending = 0;
    my $last_msg = undef;
    my $data_code = undef;
    my $datadir = undef;
    my $datafile = undef;
    my $mjd = undef;
    my $mode = undef;

    my $sel = new IO::Select($parent, $obs->socket);
    while (1) {
      while(@ready = $sel->can_read) {
	foreach my $fh (@ready) {
	  if ($fh == $obs->socket) {
	    # Message from observatory
	    if ($data_pending) { # A previous message said we would now be
	                         # getting data
	      my $status = $fh->recv($buf, $data_pending);
	      $last_msg = 1e999; # Very large

	      if (! defined $status) {
		warn "$childmsg Error reading data from observatory socket\n";
		$data_pending=0; # What else to do...
		$data_code=undef;
		$datadir = undef;
		$datafile = undef;
		$mjd = undef;
		$mode = undef;
	      } elsif (length $buf == 0) {
		# Socket must have closed
		print "Remote obsdata socket closed\n";
		exit;
	      } else { # Got data OK do somethign with it

		if ($data_pending<length($buf)) {
		  warn "$childmsg Something went really wrong with that data ".
		    "transfer\n";
		  warn "$childmsg More data that expected was received\n";
		  $data_pending=0;
		  $data_code = undef;
		  $datadir = undef;
		  $datafile = undef;
		  $mjd = undef;
		  $mode = undef;
		} else {

		  # Temporary - development
		  if ($last_msg-$data_pending>100000) {
		    #printf "$childmsg Still got %d kbytes to go\n", 
		    #  $data_pending/1024;
		    $last_msg = $data_pending;
		  }

		  $status = syswrite(OBSDATA, $buf);
		  if (! $status) {
		    warn "Error writing data file: $!\n";
		    Shutdown();
		  } elsif ($status != length($buf)) {
		    warn "Error writing data file\n";
		    Shutdown();
		  }

		  $data_pending -= length($buf);


		  if ($data_pending==0) { # Now got all data

		    printf("$childmsg Got all data\n");
		    close(OBSDATA);

		    my $message = new RtFC::Command::Data_Received($data_code,
								   $mjd,
								   $datadir,
								   $datafile,
								   $mode);

		    $status = $message->send($parent, 'Parent Process');

		    $data_code = undef;
		    $datadir = undef;
		    $datafile = undef;
		    $mjd = undef;
		    $mode = undef;
		  }
		}
	      }
	    } else {
	      my $command = recv_command($fh, undef,'observatory2');
	      if (ref $command) { # Must be a command
		if ($command->type==TRANSFER_COMMAND) {
		  if ($command->transfer_type==DATA_TRANSFER) {
		    # This is what we are after
		    $data_code = $command->code;
		    $mjd = $command->mjd;
		    $mode = $command->mode;
		    $data_pending = $command->datasize;

		    printf "Got %d bytes to read\n", $data_pending;

		    my ($day, $month, $year, $ut) = mjd2cal($mjd);
		    my ($sign, $hour, $minute, $second) =
		                           time2hms($ut, 'H', 0);

		    $datadir = sprintf("%04d%02d%02d-%02d%02d%02d",
					$year, $month, $day, $hour,
					$minute, $second);

		    $datafile = sprintf("%s.data", $obs->code);
		    open(OBSDATA, '>', "$datadir/$datafile")
		              || die "Could not open $datadir/$datafile: $!\n";
		  } else {
		    warn "Received unexpected data transfer request - ".
		      "aborting\n";
		    exit;
		  }
		} else {
		  printf STDERR "$childmsg Unknown command (%d) received\n",
		    $command->type;
		}
	      } elsif (! defined $command) { # There was a read error
		warn "$childmsg Observatory data connection lost\n";
		exit;
	      } else {
		warn "$childmsg Unexpected mesage received!!! ";
	      }
	    }
	  } elsif ($fh == $parent) {
	    # Message from parent
	    my $command = recv_command($fh, undef, 'parent');

	    if (ref $command) { # Must be a command
	      print "$childmsg: Got message from parent\n";

	      printf " of type %s\n", $command->name;

	    } else {
	      next if (! defined $command); # There was a read error
	      if ($command==-1) { # Socket was closed remotely
		warn "$childmsg Parent data connection lost\n";
		exit;
	      } else {
		warn "$childmsg Unknown error ($command) on Parent ".
		  "connection\n";
	      }
	    }

	  } else {
	    warn "$childmsg Got data from unknown socket!!\n"
	  }
	}
      }
    }
  }
}

sub setup_correlation ($) {
  my ($command) = @_;

  my ($day, $month, $year, $ut) = mjd2cal($command->mjd);
  my ($sign, $hour, $minute, $second) = time2hms($ut, 'H', 0);

  my $newcorr = $command->newcorr;
  
  my $msg;
  if ($newcorr) {
    $msg = 'Grab';
  } else {
    $msg = 'Re-correlation';
  }

  printf("$msg requested for %02d/%02d/%04d %s %.1f\n", $day, $month, $year,
	 turn2str($ut, 'H', 0), $command->integration);

  my @values = @{$command->value};
  my @connected_obs = @{$command->observatories};

  my $thiscorr = new RtFC::Correlation($command->code,
				       $command->mjd,
				       $command->integration,
				       $command->mode,
				       $corr_lags,
				       $pack,
				       \%offsets,
				       \%clockrates,
				       \@connected_obs,
				      );
  my $dir = sprintf("%04d%02d%02d-%02d%02d%02d", $year, $month, $day,
		    $hour, $minute, $second);
  if ($newcorr) {
    # Create a directory to store the data
    if (-d $dir) {
      warn "Requested to create existing directory $dir\n";
    } else {
      mkdir $dir || die "Could not create $dir: $!\n";
    }

    $Corrs{$command->code} = $thiscorr;
  } else {
    # Does the directory exist
    if (! -d $dir) {
      warn "Directory $dir does not exists. Cannot continue re-correlation\n";
      #delete $command->value->[0]; ##Not sure what this was meant to do...
      return;
    }

    do_recorr($thiscorr, $dir);
  }
}

sub do_recorr($$) {
  my ($corr, $dir) = @_;

  my @obs = keys(%{$corr->obs});
  
  foreach (@obs) {
    my $file = "$dir/${_}.data";
    if (! -f $file) {
      warn "$file does not exist. Skipping $_\n";
      next;
    }
    $corr->obs->{$_} = $file;
  }

  correlate($corr->mjd, $corr, $corr->obs, \@obs, $dir, $corr->mode, @obs);
}

sub correlate ($$$$$$@) {
  my ($mjd, $corr, $obs, $baselines, $dir, $packmode, @expected_obs) = @_;
  
  my $pwd = `pwd`;
  chomp $pwd;

  # Determine the values to write to the config file

  my ($day, $month, $year, $ut) = mjd2cal($mjd);
  my ($sign, $hour, $minute, $second) = time2hms($ut, 'H', 0);
  my $epoch = sprintf('%04d%02d%02d:%02d%02d%02d', $year, $month,
		      $day, $hour, $minute, $second);
  my $time = sprintf('%02d:%02d:%02d', $hour, $minute, $second);

  # Find which vexfile this mjd corresponds to

  my $vex;
  for (my $i=0; $i<@vextimes; $i++) {
    if ($vextimes[$i]->[0] < $mjd && $vextimes[$i]->[1] > $mjd) {
      $vex = $vex[$i];
    }
  }
  if (!defined $vex) {
    warn "Could not find schedule corresponding to $epoch\n";
    return;
  }

  # Read the vexfile for experiment id, source name and position
  # and observing frequency
  my $exper_name = $vex->exper->exper_name;

  printf("Correlating data for %s\n", $exper_name);

  my $found = 0;
  my ($sourcename, $modename, $ra, $dec, @freqs, @sideband, @pol, @bw, $rate);
  my @scans = $vex->sched;
  foreach (@scans) {
    
    last if ($_->start > $mjd);
    next if ($_->stop < $mjd);
    
    # This must be the scan we want!
    $found = 1;
    printf("Found: %s start=%s source=%-8s\n", $_->scanid,  $_->startepoch, 
	   $_->source);
    $sourcename = $_->source;
    $modename = $_->mode;
    last;
  }
  if (!$found) {
    print "No scan found\n";
    return;
  }

  my $source = $vex->source($sourcename);

  $sourcename = $source->source_name;

  my $equinox = $source->ref_coord_frame;
  $ra = rad2turn($source->ra);
  $dec = rad2turn($source->dec);

  if (uc($equinox) eq 'DATE') {
    # Do nothing as this is what we want
  } elsif (uc($equinox) eq 'B1950' || uc($equinox) eq 'J2000') {
    if (uc($equinox) eq 'B1950') {
      ($ra, $dec) = fk4fk5($ra, $dec); # B1950 -> J2000
    }
    ($ra, $dec) = J2000todate($ra, $dec, $mjd);
  } else {
    warn "Unknown reference coordinate frame ($equinox)\n";
  }
  $ra = turn2str($ra, 'H', 4);
  $dec = turn2str($dec, 'D', 3);
  
  # Get a list of all antenna we can correlate (ie the valid ones)
  
  my @obsgot;
  foreach my $o (@expected_obs) {
    if (defined $obs->{$o}) {
      push @obsgot, $o;
    }
  }
  return if (scalar(@obsgot)<2); # Only one antenna present
  
  # Just lookup freq etc for first antenna (need to generalise this)
  
  my $mode = $vex->mode($modename)->{$obsgot[0]};
  $rate = $mode->sample_rate;

  print "Mode: $modename  $rate\n";
  $rate->unit('s/sec');
  $rate = $rate->value;
  
  my @chans = $mode->chan_def;

  foreach (@chans) {
    push @freqs, $_->freq->value;
    if ($_->sideband eq 'L') {
      push @sideband, -1;
    } else {
      push @sideband, 1;
    }
    push @pol, $_->pol;
    push @bw, $_->bw->value;
  }

  my $channels = $corr->nchan;
  my $pack = $corr->pack;

  my %offsets = %{$corr->offsets};
  my %clockrates = %{$corr->clockrates};

  # Fill in default values
  foreach (@obsgot) {
    print "Doing defaults for $_\n";
    if (!exists $offsets{lc($_)}) {
      $offsets{lc($_)} = 0.0;
      print "Setting $_ clock delay to 0.0\n";
    }
    if (!exists $clockrates{lc($_)}) {
      $clockrates{lc($_)} = 0.0;
      print "Setting $_ clock rate to 0.0\n";
    }
  }

  my $ntel = scalar(@obsgot);
  my $integration = $corr->integration;

  # Need to work out how many channels in data file. First count the number of x's
  my @packmode = reverse(split //, $packmode);

  # Create the config file
  my $fconfig = "$dir/$exper_name.config";
  print "Writing $fconfig\n";
  open(CONFIG, '>', $fconfig) || die "Could not open $fconfig: $!\n";

  # Write the job specific stuff
  print CONFIG <<EOF;
NUMBER OF CHANNELS: $channels
DELAY FILENAME:     $delayfile
OUTPUT PREFIX:      ${exper_name}-
NUMBER TELESCOPES:  $ntel
NUMBER DATA FILES:  1
INTEGRATION TIME:   $integration
PRODUCT MODE:       FULL
PP FILTERBANK ON:   FALSE
OVERLAP FACTOR:     1
OUTPUT FORMAT:      ASCII
WRITE AUTOCORRS:    TRUE
PULSAR BIN ON:      FALSE
EOF

  # Correlate all antennas present so far
  my $nchan;
  my $nfreq = undef;
  my $first = 1;
  my @pols = ();
  my @plotfreq = ();
  foreach my $ant (@obsgot) {

    # Determine type of data
    my (@bw_packed, @freq_packed, @pol_packed, @side_packed);
    my $band = 0;
    foreach my $m (@packmode) {
      if ($m eq 'x') {
	push @bw_packed, $bw[$band];
	push @freq_packed, $freqs[$band];
	push @pol_packed, $pol[$band];
	if ($sideband[$band]==1) {
	  push @side_packed, 'FALSE';
	} else {
	  push @side_packed, 'TRUE';
	  }
      }
      $band++;
      last if ($band>=scalar(@freqs));
    }
    $nchan = scalar(@freq_packed);
    
    my %freqs = ();
    # Determine number of frequencie
    for (my $i=0; $i<$nchan; $i++) {
      # Count uniq sideband/freq setups
      $freqs{"$freq_packed[$i]-$side_packed[$i]"} = 1; 
    }
    my $mfreq = scalar(keys(%freqs));
    if (!defined $nfreq) {
      $nfreq = $mfreq;
    } elsif ($mfreq!=$nfreq) {
      warn "Inconsistent number for subbands per antenna!\n";
      $nfreq = $mfreq if ($mfreq>$nfreq);
    }

    # The correlator mode
    my $corrmode;
      
    # Assume all bandwidths the same
    if ($ant eq 'Hh') {
      $corrmode = 'VLBA';
    } elsif ($bw_packed[0]==64) {
      $corrmode = 'STDPACK';
    } elsif ($nchan==4 || $nchan==8) {
      if ($bw_packed[0]!=16 && $bw_packed[0]!=8) {
	warn "Unsupported correlator mode\n";
	return;
      }
      if ($pack==1) {
	$corrmode = 'STD';
      } else {
	$corrmode = 'VSOP';
      }
    } elsif ($nchan<4) {
      if ($bw_packed[0]==16 || $bw_packed[0]==8) {
	if ($pack==1) {
	  $corrmode = 'STDPACK';
	} else {
	  $corrmode = 'VSOPPACK';
	}
      } else {
	if ($pack==1) {
	  $corrmode = 'STDSUB16MHZ';
	} else {
	  $corrmode = 'VSOPSUB16MHZ';
	  }
      }
    } else {
      warn "Unsupported correlator mode!\n";
    }

    my $station = $vex->station($ant);
    my $site_name = $station->site_name;
    my $clockdelay = $offsets{lc($ant)};
    my $clockrate = $clockrates{lc($ant)};

    print CONFIG <<EOF;
TELESCOPE 1 NAME:   $site_name
MODE:               $corrmode
EOF
    if ($ant eq 'Hh') {
      print CONFIG <<EOF;
FANOUT:             4
EOF
    }

    print CONFIG <<EOF;
TSYS:               50.0
CLOCK DELAY (us):   $clockdelay
CLOCK RATE (us/s):  $clockrate
NUMBER OF BANDS:    $nchan
NUMBER OF FREQS:    $nfreq
EOF

    foreach (1..$nfreq) {
      print CONFIG<<EOF;
CLK OFFSET 0 (us):  0.0
EOF
    }

    for (my $i=0; $i<$nchan; $i++) {
      my $b = $i+1;
      print CONFIG <<EOF;
BAND $b BW (MHz):    $bw_packed[$i]
BAND $b FLIPPED:     $side_packed[$i]
BAND $b FREQ (MHz):  $freq_packed[$i]
BAND $b POL:         $pol_packed[$i]
EOF
    }

    my $datafile = "$pwd/".$obs->{$ant};
    print CONFIG <<EOF;
DATA FILE 0:        $datafile
EOF

    if ($first) {
      $first = 0;
      @pols = @pol_packed;
      @plotfreq = @freq_packed;
    }
  }
  close(CONFIG) || die "Error closing $fconfig\n";

  launch_corr($fconfig, $integration, $pwd, $dir);

  print "**POLS: @pols\n";
  print "**FREQ: @plotfreq\n";

  plotfringe($vex, $nchan, $nfreq, $dir, $sourcename, $time,  @pols, @plotfreq, @obsgot);

}

sub Initialize () {
  # Create a listening socket for the observatory connection
  # to the corr process
  # Connect to the central "server"

  my $sock_obs = IO::Socket::INET->new(LocalPort => CORRPORT,
				       Listen    => 10,
				       Resuse    => 1)
    || die "Couldn't create server socket for Observatories: $@\n";

  # Open a connection to Craigs correlator interface program
  my $sock_control = IO::Socket::INET->new(PeerAddr => $control_host,
					   PeerPort => OBSPORT
					  )
    || die "Couldn't create socket connection to control process: $@\n";

  # Register this as a correlator connection
  my $command = new RtFC::Command::Antenna_Name('CORRELATOR');
  my $status = $command->send($sock_control, 'control process');
  # Ignore return value for now

  return ($sock_obs, $sock_control);

}

sub Shutdown () {
  warn "Calling shutdown routine\n";

  foreach my $obs (@Obs) {
    $obs->socket->shutdown(2);
  }

  foreach my $obs (@ObsData) {
    $obs->socket->shutdown(2);
  }

  $sock_obs->shutdown(2) if (defined $sock_obs);
  $sock_control->shutdown(2) if (defined $sock_control);

  exit(0);
}

sub child_reaper {
  #print "Child reaper was called @_\n";
}

sub launch_corr ($$$$) {
  my ($config, $tint, $pwd, $dir) = @_;

  my $CORR_EXECUTABLE = "/home/ssi/adeller/correlator/phase1fx/fxcorr/src/fxcorr";

  # Get a list of node loads. Avoid slower nodes

  #my $gstat = `gstat -a1 -p 8651 | grep -e tera -e aqua`;
  my $gstat = `gstat -a1 -p 8651 | grep -e tera `;

  my @gstat = split /\n/, $gstat;

  my %nodes = ();
  foreach (@gstat) {
    if (/(\S+)\s+   # Node name
	 \d+\s+\(\s*\d+\/\s*\d+\)\d*\s* # ?
	 \[\s*(\S+),\s*(\S+),\s*(\S+)\]\s* # Load
	 \[\s*(\S+),\s*(\S+),\s*(\S+),\s*(\S+)\] # ?
	 \s*ON
	 /x) {

      $nodes{$1} = $2;
    } else {
      warn "Ignoring $_";
    }
  }
  my @sorted_nodes = sort {$nodes{$a} <=> $nodes{$b}} keys(%nodes);

  my $node = $sorted_nodes[0];
  
  my $exec = "rsh $node 'cd $pwd/$dir \&\& $CORR_EXECUTABLE $pwd/$config $tint'";

  print "Launching on $node (load $nodes{$node})\n";
  print "    $exec\n";

  system $exec;
}

sub plotfringe ($$$$$$\@\@@) {
  my ($vex, $nchan, $nfreq, $dir, $source, $time,  $pols, $freqs, @obs) = @_;

  my $exper = $vex->exper->exper_name;

  my @antname;
  foreach (@obs) {
    my $station = $vex->station($_);
    push @antname, $station->site_name;
  }

  my $exportpath = "/home/ssi/cphillip/public_html";

  if (! -d $exportpath) {
    warn "$exportpath does not exists!. Saving locally\n";
    $exportpath = ".";
  }

  $exportpath = "$exportpath/$exper";
  if (! -d $exportpath) {
   umask 0000;
   my $status = mkdir $exportpath;
   if (! $status) {
     warn "Could not create directory $exportpath: $!\n";
     return;
   }
 }

  my $ntel = scalar(@obs);
  for (my $i=0; $i<$ntel; $i++) {
    for (my $j=$i; $j<$ntel; $j++) {
      my $ant1 = $antname[$i];
      my $ant2 = $antname[$j];
      my @chans = ();

      if (($ant1 eq $ant2 && $nchan<=4) || $nchan<=2 
	  || ($nchan==4 && $nfreq==4)) {
	for (my $i=0; $i<$nchan; $i++) {
	  push @chans, $i;
	}

	  
	my $dev = "$exportpath/$time-$ant1-$ant2.gif/gif";

	plotit($antname[$i], $antname[$j], $exper, $dir, $dev, $time, $source,
	       $pols, $freqs, @chans);
      } else {
	# Assume RRLLRRLL channel order
	for (my $if=1; $if<=$nfreq; $if++) {
	  my $dev = "$exportpath/$time-$ant1-$ant2-IF${if}.gif/gif";
	  @chans = ($if-1, $if+1);

	plotit($antname[$i], $antname[$j], $exper, $dir, $dev, $time, $source,
	       $pols, $freqs, @chans);
	}
      }
    }
  }
}

sub plotit ($$$$$$$$$@) {
  my ($ant1, $ant2, $exper, $path, $dev, $time, $source, $pols,  $freqs, @chans) = @_;

  my ($chanlist, $polstr, $freqstr);
  foreach (@chans) {
    $chanlist .= " -c $_";
    $polstr .= " -pol $pols->[$_]";
    $freqstr .= " -freq '$freqs->[$_] MHz'";
  }

  print "****", @$freqs, "\n";

  my $cmd = "rtfc_plot.pl -ant1 $ant1 -ant2 $ant2 -exper $exper -path $path ".
    "-dev $dev -time $time -source $source $chanlist$polstr$freqstr";
  print "$cmd\n";
  system $cmd;
}
