#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include "vheader.h"

#define STRSIZE 100
#define BUFSIZE 1024*1024
#define MAXCHAN 8

void reorderbytes(char *buf, int nread, char lookup[]);

int main (int argc, char *argv[]) {
  int  opt, status, datafile, outfile, i, j, ishift, nchan, nkeep, eof;
  long bytestoread, ntoread, nwrite;
  char fname[STRSIZE+1], lookup[256];
  unsigned char *buf, *obuf, *iptr, *optr, mask1, mask2;
  unsigned int mask, submask, chanmask, *data;
  ssize_t nread;
  struct tm time, *timep;
  time_t ttime;
  encodingtype encoding;
  vhead *header;

  /* Options */
  int rate;              /* Total Sampling data rate in Mbytes/sec */
  int offset;            /* Offset in seconds from start of file we want to 
			    extract */
  float integration;     /* Amount of data in seconds we want to extract */
  char tempname[STRSIZE+1];  /* Name for temp file */
  char mode[STRSIZE+1];  /* Current mode we want to extract */
  int invert = 0;        /* Invert channels? */
  char invertchan[8];    /* Channels to invert */
  int swapped;           /* IFs swapped (need to swap nybbles */
  int reorder = 0;       /* Reorder channels? */
  int chanorder[4];      /* New order for channels */
  encodingtype defaultencoding; /* VLBA or AT bit encoding 
				   (offset binary or sign/mag) */

  struct option options[] = {
    {"mode", 1, 0, 'm'},
    {"offset", 1, 0, 'o'},
    {"integration", 1, 0, 'i'},
    {"invert", 1, 0, 'I'},
    {"tempfile", 1, 0, 'f'},
    {"outfile", 1, 0, 'f'},
    {"reorder", 1, 0, 'r'},
    {"vlba", 0, 0, 'v'},
    {"at", 0, 0, 'a'},
    {"signmag", 0, 0, 'a'},
    {"swapped", 0, 0, 'x'},
    {0, 0, 0, 0}
  };

  strcpy(mode, "xxxx");
  strcpy(tempname, "VSIPACK.tmp");
  defaultencoding = NONE;

  offset = 0;
  integration = -1;
  swapped = 0;
  
  mask1 = mask2 = 0;
  ishift = 0;

  while ((opt = getopt_long_only(argc, argv, "i:", options, NULL)) != EOF)
    switch (opt) {
    case 'm':
      if (strlen(optarg)>STRSIZE) {
	fprintf(stderr, "MODE option too long\n");
	return 0;
      }
      strcpy(mode, optarg);
      break;

    case 'I':
      if (strlen(optarg)>MAXCHAN) {
	fprintf(stderr, "Invert option (%s) too long\n", optarg);
	return 0;
      }
      strncpy(invertchan, optarg, MAXCHAN);
      invert = 1;
      break;

    case 'r':   // Reorder 2 bit channels

      if (strlen(optarg)!=4) {
	fprintf(stderr, "Must pass 4 channel reordering, e.g. \"2143\"\n");
	return 0;
      }
      // Check all only 1-4 is present and 1-4 are each present

      for (i=0; i<4; i++) {
	switch(optarg[i]) {
	  case '1':
	    chanorder[i] = 0;
	    break;
	  case '2':
	    chanorder[i] = 1;
	    break;
	  case '3':
	    chanorder[i] = 2;
	    break;
	  case '4':
	    chanorder[i] = 3;
	    break;
	default:
	  fprintf(stderr, "Cannot reorder channel %c\n", optarg[i]);
	  return 0;

	  break;
	}
      }
      reorder = 1;

      break;

    case 'o':
      status = sscanf(optarg, "%d", &offset);
      if (status !=1) {
	fprintf(stderr, "Error interpreting offset option (%s)\n", optarg);
	return(0);
      }
      break;

    case 'f':
      if (strlen(optarg)>STRSIZE) {
	fprintf(stderr, "FILENAME option too long\n");
	return 0;
      }
      strcpy(tempname, optarg);
      break;

    case 'i':
      status = sscanf(optarg, "%f", &integration);
      if (status !=1) {
	fprintf(stderr, "Error interpreting integration option (%s)\n", optarg);
	return(0);
      }
      break;

    case 'v':
      defaultencoding = VLBA;
      break;

    case 'a':
      defaultencoding = AT;
      break;

    case 'x':
      swapped = 1;
      break;

    default:
      return(0);
      break;
    }

  /* Should be just one option now */

  if (argc != optind+1) {
    fprintf(stderr, "Usage: vsipack [options] <filename>\n");
    return(0);
  }
  if (strlen(argv[optind]) > STRSIZE) {
    fprintf(stderr, "filename too long\n");
    return(0);
  } 

  if (swapped) { /* For some modes the best way of handling an IF
		     inversion is to change modes */

    if (strcasecmp(mode, "ooox")==0 || strcasecmp(mode, "ooxo")==0 ||
	strcasecmp(mode, "oxoo")==0 || strcasecmp(mode, "xooo")==0 ||
	strcasecmp(mode, "ooxx")==0 || strcasecmp(mode, "xxoo")==0) {
      char tmp;
      swapped=0;
      tmp = mode[0];
      mode[0] = mode[2];
      mode[2] = tmp;
      tmp = mode[1];
      mode[1] = mode[3];
      mode[3] = tmp;
      printf("Warning: disabling -swapped and using mode %s\n", mode);
    } else if (strcasecmp(mode, "xxxx")==0) {
      mask1 = 0xF0;
      mask2 = 0x0F;
      ishift = 4;
    } else if (strcasecmp(mode, "oxox")==0 || (strcasecmp(mode, "xoxo")==0)) {
      mask1 = 0xCC;
      mask2 = 0x33;
      ishift = 2;
    }
  }

  // Setup channel reording (if needed) 
  if (reorder) {
    for (i=0; i<256; i++) {
      lookup[i] = 0;
      for (j=0; j<4; j++) {
	lookup[i] |= ((i&(0x3<<(chanorder[j]*2)))>>chanorder[j]*2)<<j*2;
      }
    }
  }

  strcpy(fname, argv[optind]);

  datafile = open(fname, O_RDONLY|O_LARGEFILE);
  if (datafile == -1) {
    fprintf(stderr, "Error opening %s\n", fname);
    perror(NULL);
    return(0);
  }

  /* Read the header */
  header = newheader();
  readheader(header, datafile);

  // See how many channels we are dropping
  nchan = 0;
  nkeep = 0;
  for (i=0; i<strlen(mode); i++) {
    nchan++;
    if (mode[i]=='x')
      nkeep ++;
    else if (mode[i] != 'o') {
      fprintf(stderr, "Unknown compression mode %c\n", mode[i]);
    }
  }
  if (strcasecmp(mode, "xxxx")!=0 && strcasecmp(mode, "xx")!=0) {
    if (nchan != header->nchan) {
      fprintf(stderr, "Mode %s not consistent with %d channels\n", mode, 
	      header->nchan);
    }
  }
  
  mask = 0;
  if (invert) {
    if (defaultencoding==NONE)
      encoding = header->encoding;
    else
      encoding = defaultencoding;

    if (encoding == VLBA) {
      chanmask = 0x3; /* 11 */
    } else {  /* AT */
      chanmask = 0x1; /* 01 */ 
    }

    if (8%nchan !=0) {
      fprintf(stderr, "%d channels does not fit into 32bit word. Cannot invert\n", 
	      nchan);
      return(0);
    }

    // Build up submask for all channels
    submask = 0x0;
    for (i=0; i<nchan; i++) {
      if (i<strlen(invertchan) && invertchan[strlen(invertchan)-1-i]=='x')
	submask |= (chanmask&0x3) << i*2;
    }

    for (i=0; i*2*nchan*2<32; i++) {
      mask |= submask << i*2*nchan*2; // Assumes 2 bit
    }

    printf("Using invert mask %x\n", mask);

  }


  outfile = creat(tempname, S_IREAD|S_IWRITE|S_IRGRP);
  if (outfile == -1) {
    fprintf(stderr, "Error creating output file\n");
    perror(NULL);
    return(0);
  }

  /* Calculate the data rate, assuming Nyquist */
  rate = header->nchan*header->bandwidth*header->numbits*2;

  /* Get the current time and add the offset */
  gettime(header, &time);

  ttime = mktime(&time);
  ttime += offset;
  timep = localtime(&ttime);
  
  if (timep->tm_isdst) { /* Adjust for DST oddness */
    ttime -= 60*60;
    timep = localtime(&ttime);
  }
  settime(header, timep);

  /* Reduce the number of channels appropriately */
  header->nchan = header->nchan * nkeep/nchan;

  writeheader(header, outfile, NULL);

  /* First position ourselves at the start of valid data */
  status = lseek(datafile, offset*rate*1e6/8, SEEK_CUR);
  if (status==-1) {
    fprintf(stderr, "Error moving to start of required data (offset too large?)\n");
    return(0);
  }

  buf = malloc(BUFSIZE);
  if (buf==NULL) {
    fprintf(stderr, "Error allocating memory\n");
    return(0);
  }


  if (strcasecmp(mode, "xxxx")==0 || strcasecmp(mode, "xx")==0) {
    /* Default = copy everything */
    obuf = buf;
  } else {
    obuf = malloc(BUFSIZE);
    if (obuf==NULL) {
      fprintf(stderr, "Error allocating memory\n");
      return(0);
    }
  }

  if (integration>0) 
    bytestoread = integration*rate*1e6/8;
  else 
    bytestoread = -1;

  eof = 0;
  ntoread = BUFSIZE;
  while (bytestoread!=0) {

    if (bytestoread>0 && bytestoread<ntoread) ntoread = bytestoread;

    nread = read(datafile, buf, ntoread);

    if (nread==-1) {
      fprintf(stderr, "Error reading from %s\n", fname);
      perror(NULL);
      return(0);
    } else if (nread==0) {
      eof = 1;
      break;
    } 
    bytestoread -= nread;

    if (invert) {
      /* Invert the spectrum. Do this by XORing with a mask */
      data = (unsigned int*)buf;
      for (i=0; i<nread; i+=4) {
	*data = *data ^ mask;  
	data++;
      }
    }

    if (reorder) reorderbytes(buf, nread, lookup);

    /* OK we have the data - unpack it */

    if (strcasecmp(mode, "xxxx")==0 || strcasecmp(mode, "xx")==0) {
      /* Save all data - nothing to do */
      nwrite = nread;

    } else if (strcasecmp(mode, "ooox")==0) {
      /* Just first channel */
      
      if (nread%4 !=0) {
	fprintf(stderr, "Oops, I expect to read even number of bytes, aborting\n");
	close(outfile);
	return(0);
      }
      iptr = buf;
      optr = obuf;
      for (i=0; i<nread/4; i++) {
	*optr = (*iptr & 0x3);
	iptr++;
	*optr |= (*iptr & 0x3) <<2;
	iptr++;
	*optr |= (*iptr & 0x3) <<4;
	iptr++;
	*optr |= (*iptr & 0x3) <<6;
	iptr++;
	optr++;
      }
      nwrite = nread/4;

    } else if (strcasecmp(mode, "ooxo")==0) {
      /* Just first channel */
      
      if (nread%4 !=0) {
	fprintf(stderr, "Oops, I expect to read even number of bytes, aborting\n");
	close(outfile);
	return(0);
      }
      iptr = buf;
      optr = obuf;
      for (i=0; i<nread/4; i++) {
	*optr = (*iptr & 0x0C)>>2;
	iptr++;
	*optr |= (*iptr & 0x0C);
	iptr++;
	*optr |= (*iptr & 0x0C) <<2;
	iptr++;
	*optr |= (*iptr & 0x0C) <<4;
	iptr++;
	optr++;
      }
      nwrite = nread/4;


    } else if (strcasecmp(mode, "oxoo")==0) {
      /* Just first channel */
      
      if (nread%4 !=0) {
	fprintf(stderr, "Oops, I expect to read even number of bytes, aborting\n");
	close(outfile);
	return(0);
      }
      iptr = buf;
      optr = obuf;
      for (i=0; i<nread/4; i++) {
	*optr = (*iptr & 0x30)>>4;
	iptr++;
	*optr |= (*iptr & 0x30)>>2;
	iptr++;
	*optr |= (*iptr & 0x30);
	iptr++;
	*optr |= (*iptr & 0x30)<<2;
	iptr++;
	optr++;
      }
      nwrite = nread/4;

    } else if (strcasecmp(mode, "xooo")==0) {
      /* Just first channel */
      
      if (nread%4 !=0) {
	fprintf(stderr, "Oops, I expect to read even number of bytes, aborting\n");
	close(outfile);
	return(0);
      }
      iptr = buf;
      optr = obuf;
      for (i=0; i<nread/4; i++) {
	*optr = (*iptr & 0xC0)>>6;
	iptr++;
	*optr |= (*iptr & 0xC0)>>4;
	iptr++;
	*optr |= (*iptr & 0xC0)>>2;
	iptr++;
	*optr |= (*iptr & 0xC0);
	iptr++;
	optr++;
      }
      nwrite = nread/4;

    } else if (strcasecmp(mode, "ooxx")==0) {
      /* Just 2 adjacent channels . VSOP modes*/

      if (nread%2 !=0) {
	fprintf(stderr, "Oops, I expect to read even number of bytes, aborting\n");
	close(outfile);
	return(0);
      }
      iptr = buf;
      optr = obuf;
      for (i=0; i<nread/2; i++) {
	*optr = *iptr & 0xF;
	iptr++;
	*optr |= (*iptr & 0xF) <<4 ;
	iptr++;
	optr++;
      }
      nwrite = nread/2;
    } else if (strcasecmp(mode, "xxoo")==0) {
      /* Just 2 adjacent channels VSOP modes */

      if (nread%2 !=0) {
	fprintf(stderr, "Oops, I expect to read even number of bytes, aborting\n");
	close(outfile);
	return(0);
      }
      iptr = buf;
      optr = obuf;
      for (i=0; i<nread/2; i++) {
	*optr = (*iptr & 0xF0)>>4;
	iptr++;
	*optr |= (*iptr & 0xF0) ;
	iptr++;
	optr++;
      }
      nwrite = nread/2;

    } else if (strcasecmp(mode, "oxox")==0) {
      /* Just 2 separate channels Dual pol mode*/

      if (nread%2 !=0) {
	fprintf(stderr, "Oops, I expect to read even number of bytes, aborting\n");
	close(outfile);
	return(0);
      }
      iptr = buf;
      optr = obuf;
      for (i=0; i<nread/2; i++) {
	*optr = (*iptr & 0x03) | (*iptr & 0x30)>>2;
	iptr++;
	*optr |= (*iptr & 0x03)<<4 | (*iptr&0x30)<<2;
	iptr++;
	optr++;
      }
      nwrite = nread/2;

    } else if (strcasecmp(mode, "xoxo")==0) {
      /* Just 2 separate channels Dual pol mode*/

      if (nread%2 !=0) {
	fprintf(stderr, "Oops, I expect to read even number of bytes, aborting\n");
	close(outfile);
	return(0);
      }
      iptr = buf;
      optr = obuf;
      for (i=0; i<nread/2; i++) {
	*optr = (*iptr & 0x0C)>>2  | (*iptr&0xC0)>>4;
	iptr++;
	*optr |= (*iptr & 0x0C)<<2 | (*iptr&0xC0);
	iptr++;
	optr++;
      }
      nwrite = nread/2;

    } else if (strcasecmp(mode, "xo")==0 || strcasecmp(mode, "ox")==0) {
      /* 16 bit mode, drop every 2nd byte*/

      if (nread%2 !=0) {
	fprintf(stderr, "Oops, I expect to read even number of bytes, aborting\n");
	close(outfile);
	return(0);
      }
      iptr = buf;
      optr = obuf;

      if (strcasecmp(mode, "xo")==0) iptr++;

      for (i=0; i<nread/2; i++) {
	*optr = *iptr;
	optr++;
	iptr+=2;
      }
      nwrite = nread/2;
    } else {
      fprintf(stderr, "Unsupported mode %s\n", mode);
      nwrite = 0;
      return(0);
    }

    /* Swap IFs if need be. Could make this as part of the packing,
       at the expense of code noise */
    if (swapped) {
      unsigned char tmp;
      iptr = obuf;
      for (i=0; i<nwrite; i++) {
	tmp = ((*iptr<<ishift)&mask1) | ((*iptr>>ishift)&mask2);
	*iptr = tmp;
	iptr++;
      }
    }

    status = write(outfile, obuf, nwrite);
    if (status != nwrite) {
      fprintf(stderr, "Error writing to output file\n");
      perror(NULL);
      return(0);
    }
  }
  
  free(buf);
  if (strcasecmp(mode, "xxxx")!=0 && strcasecmp(mode, "xx")!=0)
    free(obuf);
  close(datafile);
  close(outfile);
  
  if (integration>0 && bytestoread>0) {
    fprintf(stderr, "Hit EOF before reading %.1f seconds\n", integration);
    return(0);
  }
  return(1);
}

void reorderbytes(char *buf, int nread, char lookup[]) {
  int i;
  unsigned char *ptr;

  ptr = buf;
  for (i=0; i<nread; i++) {
    *ptr = lookup[*ptr];
    ptr++;
  }
}
