// ss.cc -- TCP/IP socker server
// $Id: ss.cc,v 1.12 2006/09/15 12:36:26 lindi Exp $
// Copyright (C) 2000 Ari Mujunen. (Ari.Mujunen@hut.fi)

// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License.
// See the file 'COPYING' for details.

// $Log: ss.cc,v $
// Revision 1.12  2006/09/15 12:36:26  lindi
// common: use SO_REUSEADDR so that tcp port can be reused immediately.
//
// Revision 1.11  2005/11/30 10:14:27  lindi
// If read() from network returns multiple lines parse them separately
//
// Revision 1.10  2004/05/10 07:16:17  amn
// Removed IP address check.
//
// Revision 1.9  2002/12/18 07:31:25  amn
// Upped revision.
//
// Revision 1.2  2002/12/18 07:31:23  amn
// intermediate
//
// Revision 1.1  2002/12/18 07:23:46  amn
// Recovered from /home disk crash by re-creating 'common' module in CVS.
//
// Revision 1.8  2000/02/23 13:08:22  amn
// 'write()' error return conditions were just the opposite.
//
// Revision 1.7  2000/02/23 13:01:07  amn
// Asserts away from "UNknown command" network replies.
//
// Revision 1.6  2000/02/23 12:09:53  amn
// Added ignoring all SIGPIPE signals during 'serve_socket()' loop.
// Read/write calls will return EPIPE errors instead.
//
// Revision 1.5  2000/02/22 04:40:27  amn
// Removed 'exit()'s from the inside of serve_socket loop.
//
// Revision 1.4  2000/02/16 11:59:37  amn
// Removed connect/disconnect messages (stderr).
//
// Revision 1.3  2000/02/09 07:36:47  amn
// In case of default "unknown keyword" reply, send back a terminating '\n'.
//
// Revision 1.2  2000/02/04 13:55:49  amn
// Added networking include files.
//
// Revision 1.1  2000/01/17 14:26:50  amn
// Initial version.
//

#include "registration.h"
RCSID(ss_cc, "$Id: ss.cc,v 1.12 2006/09/15 12:36:26 lindi Exp $");


// Socket call includes.
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>  // for inet_ntoa()

#include <signal.h>  // for ignoring SIGPIPE


//----------------------------------------------------------------------
class cMessageProcessor {
public:
  virtual int access_level(char *ip_address_string);
  virtual int process_message(int filedes);
  virtual int process_message_buffer(int filedes, char *buffer, int nbytes);
  virtual void doBackground(void);
};


#define MAXMSG	512

// const char our_ip_addresses[] = "193.166.42.";
// Allow all:
const char our_ip_addresses[] = "";

int cMessageProcessor::access_level(char *ip_address_string)
{
  // Compare only the leading characters of allowed own addresses.
  int len = strlen(our_ip_addresses);

  // If the incoming IP address begins with the allowed string, then ok.
  if (strncmp(ip_address_string, our_ip_addresses, len) == 0) {
    return 2;  // 2 == r/w, 1=r, <1, no access, close connection
  }
  return 0;
}  // access_level

int
cMessageProcessor::process_message (int filedes)
{
  char buffer[MAXMSG];
  char *linestart, *lineseparator;
  int nbytes, ret;

  nbytes = read (filedes, buffer, MAXMSG-1);  // space for '\0'
  if (nbytes < 0) {
    // Read error.
    // xxx: I once saw 'Broken pipe' on './dammer' virtual console...
    //      Could we then return '-1' EOF???
    perror (MY_NAME "read");
    // exit (EXIT_FAILURE);
    // Now testing xxx:
    fprintf (stderr, MY_NAME "returning EOF from fd %d\n", filedes);
    return -1;
  } else if (nbytes == 0) {
    // End-of-file.
    return -1;
  } else {
    // Data read.
    buffer[nbytes] = '\0';

    // xxx: Should actually process '\n' characters or some other means
    // of detecting individual messages in the incoming stream.

#if 1
    linestart = buffer;
    while ((lineseparator = strchr(linestart, '\n')) != NULL) {
	    *lineseparator = '\0';
	    ret = process_message_buffer(filedes, linestart, lineseparator - linestart + 1);
	    if (ret != 0)
		    return ret;
	    linestart = lineseparator + 1;
    }
    return 0;
#endif
#if 0
    fprintf (stderr, MY_NAME "got message: `%s'\n", buffer);
#endif

    return process_message_buffer(filedes, buffer, nbytes);

  }
}  // process_message

int
cMessageProcessor::process_message_buffer(int filedes,
					  char *buffer,
					  int nbytes)
{
#define NWORDS 10
  char *words[NWORDS];
  const char delims[] = " \t\n\r";  // xxx: may add ",="???
  int number_of_words = 0;
  char *tok;
  char *running;
  int retCode;

  if (nbytes < 1) {
    return 0;  // null request, processed okay
  }

  buffer[nbytes-1] = '\0';  // ensure that 'buffer' ends as a string
  // Make a copy since 'strsep()' modifies its buffer (puts in '\0' chars).
  assert( (running = strdup(buffer)) != NULL );
  while ( (tok = strsep(&running, delims)) ) {
    if (number_of_words < NWORDS) {
      words[number_of_words] = tok;
      number_of_words++;
    }
  }  // while tokens available

  if (number_of_words < 1) {
    retCode = 0;  // null request, processed okay
    goto lExit;
  }

  if (strcmp(words[0], "quit")==0) {
    retCode = -1;  // end-of-file, close connection
    goto lExit;
  }
  if (strcmp(words[0], "end")==0) {
    retCode = -1;  // end-of-file, close connection
    goto lExit;
  }
  if (strcmp(words[0], "exit")==0) {
    retCode = -1;  // end-of-file, close connection
    goto lExit;
  }
  if (strcmp(words[0], "GET")==0) {
    // Default reply to an HTTP message.
    const char msg[] =
      // xxx: actually this is in incoming 'GET' message, not in reply:
      // "Content-type: text/html\n\n"
      "<HEAD><TITLE>404 Not Found</TITLE></HEAD>\n"
      "<BODY><H1>404 Not Found</H1>"
      "The requested URL was not found on this server.<P>\n"
      "</BODY>";
    assert ( write (filedes, msg, sizeof(msg)-1) == (sizeof(msg)-1) ); 
      // don't write trailing '\0'
    retCode = -1;  // close connection
    goto lExit;
  }

  // "Unknown message":
  // Sometimes the other end has gone away before we get the chance
  // to write back this message.
  if ( write (filedes, "UN,", 3) != 3 ) {
    perror (MY_NAME "write");
    fprintf (stderr, MY_NAME "writing UN error reply message to fd %d\n", filedes);
    retCode = -1;  // close (broken) connection
    goto lExit;
  }
  if ( write (filedes, buffer, nbytes) != nbytes ) {
    perror (MY_NAME "write");
    fprintf (stderr, MY_NAME "writing UNknown '%*s' to fd %d\n", nbytes, buffer, filedes);
    retCode = -1;  // close (broken) connection
    goto lExit;
  }
  if ( write (filedes, "\n", 1) != 1 ) {
    perror (MY_NAME "write");
    fprintf (stderr, MY_NAME "writing LF to fd %d\n", filedes);
    retCode = -1;  // close (broken) connection
    goto lExit;
  }
  retCode = 0;

lExit:
  if (running) {
    free(running);
  }
  return retCode;
}  // process_message_buffer

void cMessageProcessor::doBackground(void) {
  // Nothing yet.
}  // doBackground


//----------------------------------------------------------------------
class cSocketServer {
public:
  void serve_socket(int port_number, class cMessageProcessor& mp);
};


static int
make_socket (unsigned short int port)
{
  int sock, one = 1;
  struct sockaddr_in name;

  // Create the socket.
  sock = socket (PF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
    perror (MY_NAME "socket");
    exit (EXIT_FAILURE);
  }

  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(one)) < 0) {
    perror("setsockopt");
    exit (EXIT_FAILURE);
  }

  // Give the socket a name.
  name.sin_family = AF_INET;
  name.sin_port = htons (port);
  name.sin_addr.s_addr = htonl (INADDR_ANY);
  if (bind (sock, (struct sockaddr *) &name, sizeof (name)) < 0) {
    perror (MY_NAME "bind");
    exit (EXIT_FAILURE);
  }

  return sock;
}  // make_socket

void
cSocketServer::serve_socket (int port_number, class cMessageProcessor& mp)
{
  void (*old_sigpipe_handler)(int) = NULL;
  int sock;
  fd_set active_fd_set, read_fd_set;
  int sfd;

  // From libc Info file:
  // "Every library function that returns this error code also
  // generates a `SIGPIPE' signal; this signal terminates the program
  // if not handled or blocked."
  // After ignoring we'll only get EPIPE from library calls.
  assert( (old_sigpipe_handler = signal(SIGPIPE, SIG_IGN)) != SIG_ERR );

  // Create the socket and set it up to accept connections.
  sock = make_socket (port_number);
  if (listen (sock, 1) < 0) {
    perror (MY_NAME "listen");
    exit (EXIT_FAILURE);
  }

  // Initialize the set of active sockets.
  FD_ZERO (&active_fd_set);
  FD_SET (sock, &active_fd_set);

  while (1) {
    int numOfDescr;
    struct timeval timeout;

    // Block until input arrives on one or more active sockets.
    read_fd_set = active_fd_set;
    timeout.tv_sec = 0;
    timeout.tv_usec = 500000;  // 0.5 sec
    if ( (numOfDescr = select (FD_SETSIZE, &read_fd_set, NULL, NULL, &timeout)) < 0 ) {
      // xxx: Could this be another place for 'Broken pipe'???
      perror (MY_NAME "select");
#if 0
      // If numOfDescr < 0, it is ok to continue the loop.
      exit (EXIT_FAILURE);
#endif
    }

    // Service all the sockets with input pending.
    if (numOfDescr > 0) {
      for (sfd = 0; sfd < (int)FD_SETSIZE; ++sfd) {

        // Skip file descriptors with no data pending.
        if (!FD_ISSET (sfd, &read_fd_set)) {
          continue;
        }

        if (sfd == sock) {
          // Connection request on original socket.

          struct sockaddr_in clientname;
          size_t size;
          int new_sock;

          size = sizeof (clientname);
          new_sock = accept (sock,
                             (struct sockaddr *) &clientname,
                             &size);
          if (new_sock < 0) {
            perror (MY_NAME "accept");
#if 0
            // Instead of exiting we just continue the next loop run.
            exit (EXIT_FAILURE);
#else
            continue;
#endif
          }
#if DEBUG_CONNECTIONS
          fprintf (stderr, MY_NAME "connect from host %s, port %hd, (fd %d), ",
                   inet_ntoa (clientname.sin_addr),
                   ntohs (clientname.sin_port),
                   new_sock);
#endif
          // Connecting from this IP address allowed?
          if (mp.access_level(inet_ntoa (clientname.sin_addr)) < 1) {
#if DEBUG_CONNECTIONS
            fprintf (stderr, "rejected.\n");
#endif
            (void)close (new_sock);
          } else {
#if DEBUG_CONNECTIONS
            fprintf (stderr, "accepted.\n");
#endif
            FD_SET (new_sock, &active_fd_set);
          }
        } else {
          // Data arriving on an already-connected socket.

          if (mp.process_message (sfd) < 0) {

            // xxx: Some way to exit from this loop and terminate the server?
            //      (close all connections first)
            // xxx: Detecting 'reload' => close all connections? (necessary?)

            (void)close (sfd);
            // xxx: if 'sfd' is "broken pipe", is this allowed?
            FD_CLR (sfd, &active_fd_set);
#if DEBUG_CONNECTIONS
            fprintf (stderr, MY_NAME "disconnected fd %d.\n",
                     sfd);
#endif
          }
        }  // else data arriving from an existing connection
      }  // for each file descriptor
    }  // if any file descriptor is ready

    // Now (regardless if we serviced a network request or
    // if only a simple timeout occurred) we process one pending
    // background/logging request.
    mp.doBackground();

  }  // while forever

  // xxx: Most probably won't ever get executed, but anyhow...:
  assert( signal(SIGPIPE, old_sigpipe_handler) != SIG_ERR );
}  // serve_socket
