/* fixmk5crc.c -- Zero out Mk4 frame rack number LSB (just preceding
                  sync word) for Mk5A I/O board, correcting CRC12. */

/*
 * $Log: fixmk5crc.c,v $
 * Revision 1.1  2003/03/18 10:33:50  amn
 * Moved zeroing rack number LSB from 'mrg' into 'fixmk5crc'.
 *
 *
 */

/* Copyright (C) 2001--2003 Ari Mujunen, Ari.Mujunen@hut.fi

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */


/* This source gets 64/32/16/8 bit datatype definition from Makefile, */
/* so that it can be automatically built into multiple binaries. */
/* #define DATATYPE xxx */
/* #define PRINTFDATATYPE "xxx" */

/* Ensure 64-bit file support regardless of '-D_LARGEFILE64_SOURCE=1'
   in Makefile. */
#define _LARGEFILE64_SOURCE 1  /* Large File Support (LFS) '*64()' functions. */ 
#define _FILE_OFFSET_BITS 64  /* Automatic '*()' --> '*64()' replacement. */ 

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include <sys/types.h>  /* open() */
#include <sys/stat.h>  /* open() */
#include <fcntl.h>  /* open() */

#include <sys/time.h>  /* gettimeofday() */
#include <unistd.h>  /* gettimeofday(), usleep() */

#include <string.h>  /* strstr() */


typedef unsigned char tByte, *ptByte;

/* Data words, 64/32/16/8 bits. */
typedef DATATYPE tD, *ptD;


void
crc12(ptD cr, ptD d, int datawords) 
{
  int i;

  /* 'cr' CRC register is assumed to be initialized */
  /* with 12 data words. */
  /* (Init to all zero; or use old regs to continue. */
  for (i=0; i<datawords; i++) {
    tD n0  = *d++ ^ cr[11];  /* x^12 + */
    tD n11 = n0 ^ cr[10];  /* x^11 + */
    tD n3  = n0 ^ cr[2];   /* x^3 + */
    tD n2  = n0 ^ cr[1];   /* x^2 + */
    tD n1  = n0 ^ cr[0];   /* x + 1 */

    cr[11] = n11;
    cr[10] = cr[ 9];
    cr[ 9] = cr[ 8];
    cr[ 8] = cr[ 7];
    cr[ 7] = cr[ 6];
    cr[ 6] = cr[ 5];
    cr[ 5] = cr[ 4];
    cr[ 4] = cr[ 3];
    cr[ 3] = n3;
    cr[ 2] = n2;
    cr[ 1] = n1;
    cr[ 0] = n0;
  }  /* for every data word (with multiple bits) in input data stream */
  /* The result is in 'cr[]', MSB in last data word. */
}  /* crc16 */

/* Mark IV frame header in byte chunks, without parity bits. */
typedef struct sRawMk4Header {
  /* 64-bit/8-byte aux field */
  tByte auxhs1[2];
  tByte auxhs2[2];
  tByte auxtrk;
  tByte auxsrc;
  tByte auxsts;
  tByte auxsid;
  /* 4-byte sync word */
  tByte sync[4];
  /* 64-bit/8-byte time code (52 bits) plus CRC (12 bits) field */
  tByte timec[5];  /* to full seconds */
  tByte timef[3];  /* sub-second plus CRC-12 */
} tRawMk4Header;


void
mk4crc(int framecount, ptD target, ptD crcBlock)
{
  int i, j;

  /* Init CRC generator to all zeroes. */
  memset(crcBlock, 0, 12*sizeof(tD));

  /* Headstack positions, 4 bytes, 32 bits, 0--31. */

  /* Track information, 2 bytes, 16 bits, 32-47. */

  /* Formatter status, "system ID", zero LSB; 2 bytes, 16 bits, 48--63. */
  target[63] = 0;

  /* Sync word, 4 bytes, 32 bits, 64--95. */
  for (i=64; i<96; i++) {
    if (target[i] != (DATATYPE)(~(0))) {
      fprintf(stderr, "Fr# %10d: sync word bit[%d] == " PRINTFDATATYPE "\n",
              framecount, i, target[i]);
    }
  }  /* for each bit in sync word */

  /* Timecode, 6.5bytes, 52 bits, 96--147. */

  /* Re-calc CRC12. */
  crc12(crcBlock, target, 148);

  /* CRC12, 1.5bytes, 12 bits, 148--159. */
  for (i=11,j=148; i>=0; i--,j++) {
#if 0
    if (crcBlock[i] != target[j]) {
      fprintf(stderr, "Fr# %10d: crcCalc[%02d] == " PRINTFDATATYPE ", crcData[%02d] == " PRINTFDATATYPE "\n",
              framecount, i, crcBlock[i], j, target[j]);
    }  /* if calculated different than what was in data */
#endif
    target[j] = crcBlock[i];
  }  /* for each bit in crc12 */

  return;
}  /* mk4crc */



int
main(int argc, char **argv)
{
  int useParity;
  size_t framesize;
  ssize_t framebytes;
  ssize_t nread;
  ptD frame;
  ptD crcBl;
  int framecount;

  useParity = 0;
  framesize = 2500*(useParity?9:8);
  assert( (frame = calloc(framesize, sizeof(tD))) != NULL );
  framebytes = framesize * sizeof(tD);
  assert( (crcBl = calloc(12, sizeof(tD))) != NULL );

  framecount = 0;
  for (;;) {
    nread = read (STDIN_FILENO, frame, framebytes);
    if (nread < framebytes) {
      if (nread < 0) {
        perror("reading standard input");
        exit (EXIT_FAILURE);
      }
      if (nread == 0) {
        goto lEnd;
      }
      assert( (nread == framebytes) );
    }
    framecount++;
    mk4crc(framecount, frame, crcBl);
    assert( (write(STDOUT_FILENO, frame, framebytes) == framebytes) );
  }  /* forever */

lEnd:
  exit (EXIT_SUCCESS);
}  /* main */
