/* ft2.c -- Quick FFT transform of ADS-1000 1-bit 1Gbit/s sampled data. */

/*
 * $Log: ft2.c,v $
 * Revision 1.2  2002/11/29 10:58:40  amn
 * Unified ft2 and ft3; made mrg seq. numbered file opening work; wr parms.
 *
 * Revision 1.1  2002/11/01 13:19:58  amn
 * Just before changing to use a common tstamp.cc file.
 *
 */

/* Copyright (C) 2002 Ari Mujunen, Ari.Mujunen@hut.fi

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */


#include <sys/types.h>  /* for open()ing */
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>  /* for read()ing */
#include <assert.h>

#include <rfftw.h>
#include <math.h>  /* for M_PI, atan2() */


int
main(int argc, char **argv)
{
  int blksize;  /* memory / I/O block in traditional bytes */
  typedef unsigned long int tVsi;
  int blkwords;  /* mem / I/O in 32-bit words */
  int fftsize;
  int ffts_per_block;
  int totalblks;  /* how many blocks to read from file */
  char *memblk;  /* file I/O pointer */

  fftw_real *in, *out;  /* libfftw in and out arrays, sized by 'fftsize' */
  fftw_real *power_spectrum, *phase;  /* out-->pwr/phase, sized by fftsize/2+1 */
  rfftw_plan p;
  int fid;  /* opened input file number */
  int k;

  if (argc < 5) {
    fprintf(stderr, "%s: needs at least blocksize fftsize|0 totalblocks path\n", argv[0]);
    exit(EXIT_FAILURE);
  }

  /* File read block size and the corresponding memblk of 32-bit words. */
  blksize = atoi(argv[1]);
  if (blksize % sizeof(tVsi) != 0) {
    fprintf(stderr, "%s: blocksize has to be a multiple of %d\n", argv[0], sizeof(tVsi));
    exit(EXIT_FAILURE);
  }
  blkwords = blksize / sizeof(tVsi);

  assert( (memblk = malloc(blksize)) != NULL );

  /* FFT bins. */
#define SAMPLES_IN_VSI_WORD (32)
  fftsize = atoi(argv[2]);
  if (fftsize % SAMPLES_IN_VSI_WORD != 0) {
    fprintf(stderr, "%s: fftsize has to be a multiple of %d\n", argv[0], SAMPLES_IN_VSI_WORD);
    exit(EXIT_FAILURE);
  }
  if (fftsize < SAMPLES_IN_VSI_WORD) {
    /* Calculate default, the whole memblk to one FFT transform. */
    fftsize = blkwords * SAMPLES_IN_VSI_WORD;
    fprintf(stderr, "%s: defaulting fftsize to %d\n", argv[0], fftsize);
  }
  if (((blkwords * SAMPLES_IN_VSI_WORD) % fftsize) != 0) {
    fprintf(stderr, "%s: blocksize (in bits %d) has to be a multiple of fftsize (%d)\n", argv[0], (blkwords * SAMPLES_IN_VSI_WORD), fftsize);
    exit(EXIT_FAILURE);
  }
  ffts_per_block = (blkwords * SAMPLES_IN_VSI_WORD) / fftsize;
  assert( (in = (fftw_real *)calloc(fftsize, sizeof(fftw_real))) != NULL );
  assert( (out = (fftw_real *)calloc(fftsize, sizeof(fftw_real))) != NULL );
  assert( (power_spectrum = (fftw_real *)calloc(fftsize/2+1, sizeof(fftw_real))) != NULL );
  assert( (phase = (fftw_real *)calloc(fftsize/2+1, sizeof(fftw_real))) != NULL );

  totalblks = atoi(argv[3]);

  p = rfftw_create_plan(fftsize, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE);

  /* assert( (fid = open(argv[4], O_RDONLY)) != -1 ); */
  /* Try large file support. */
  assert( (fid = open64(argv[4], O_RDONLY)) != -1 );
  
  for (k=0; k<totalblks; k++) {
    tVsi *memwords;  /* memory 32-bit word pointer (starts at memblk) */
    int f, w;

    /* Read one block to memory. */
    assert( (read(fid,
                  memblk,
                  blksize)) == blksize );

    memwords = (tVsi *)memblk;
    for (f=0; f<ffts_per_block; f++) {

      fftw_real *inp = in;
      int b;

      /* 1-bit ADS-1000 sampling, LSB first in time. */
      for (w=0; w<(fftsize/SAMPLES_IN_VSI_WORD); w++) {
        *inp++ = (*memwords & 0x00000001) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00000002) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00000004) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00000008) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00000010) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00000020) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00000040) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00000080) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00000100) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00000200) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00000400) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00000800) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00001000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00002000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00004000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00008000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00010000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00020000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00040000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00080000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00100000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00200000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00400000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x00800000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x01000000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x02000000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x04000000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x08000000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x10000000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x20000000) ? 1.0 : -1.0;
        *inp++ = (*memwords & 0x40000000) ? 1.0 : -1.0;
        *inp++ = (*memwords++ & 0x80000000) ? 1.0 : -1.0;
      }  /* for each word of one FFT */
      /* FFT. */
      rfftw_one(p, in, out);
      power_spectrum[0] = out[0]*out[0];  /* DC component */
      /** phase[0] = 0.0; **/
      for (b = 1; b < (fftsize+1)/2; ++b) {  /* (b < N/2 rounded up) */
        power_spectrum[b] += out[b]*out[b] + out[fftsize-b]*out[fftsize-b];
        /** phase[b] = atan2(out[fftsize-b], out[b])*360.0/M_PI; **/
      }
      if (fftsize % 2 == 0) {  /* fftsize is even */
        power_spectrum[fftsize/2] += out[fftsize/2]*out[fftsize/2];  /* Nyquist freq. */
        /** phase[fftsize/2] = 0.0; **/
      }
    }  /* for a single FFT transform */
  }  /* for every block in input file */

  assert( (close(fid) != -1) );
      
  /* Output spectrum. */
  for (k=0; k<(fftsize/2); k++) {
    printf("%f\t%f\t%f\n", 1024.0*(double)k/(double)fftsize, power_spectrum[k], phase[k]);
  }

  rfftw_destroy_plan(p);

  return (EXIT_SUCCESS);
}  /* main */
