/* ft3summing.c -- Quick FFT transform of BBC03 LSB sampled+formatted data. 
 *
 * This is basically identical to ft3.c for data the decoding part, however
 * the FFT is performed once on the summed-up input data (O n*log(n)) instead 
 * of summing FFT results (O n^2*log(n)). Like in ft3.c, the output power
 * spectrum is the sum of all power spectra, not the average.
 *
 */

/*
 * $Log: ft3summing.c,v $
 * Revision 1.1  2006/04/28 07:31:11  jwagner
 * Same as ft3 but faster as it performs only one FFT. Coding style edited a bit, tweaked MODE_2BIT_TEST.
 *
 * Revision 1.6  2006/04/06 11:34:18  amn
 * Updated to do 2-bit VLBA samples in VSIC mode 1 and VSIB mode 7.
 *
 * Revision 1.5  2005/01/19 07:19:02  amn
 * Dec-2004 JIVE software formatter changes; vsib.c seek support.
 *
 * Revision 1.3  2004/08/02 18:35:39  amn
 * First release of sf.c, software Mk4/VLBA formatter.
 *
 * Revision 1.2  2003/03/05 08:01:56  amn
 * March 2003 Jb2/Mh Mk4/5-style Gbit test version.
 *
 * Revision 1.1  2002/11/29 10:58:40  amn
 * Unified ft2 and ft3; made mrg seq. numbered file opening work; wr parms.
 *
 *
 */

/* Copyright (C) 2002 Ari Mujunen, Ari.Mujunen@hut.fi

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */


#include <sys/types.h>  /* for open()ing */
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>  /* for read()ing */
#include <assert.h>

#include <rfftw.h>
#include <math.h>  /* for M_PI, atan2() */


#define MODE_2BIT_TEST        1     /* VLBA sampler, VSIC mode 1, VSIB mode 7, "2-bit test". */
#define MODE_BBC_3LSB         2     /* 3rd VC/BBC lower sideband, VC05 JB, BBC03 Mh. */
#define MODE_BBC_2LSB         3     /* 2nd VC/BBC lower sideband, VC03 JB, BBC02 Mh. */
#define MODE_16B_RAW_VLBA     4     /* 16-bit raw VLBA sampler data, BBC01 USB in two LSBits of 16b.w. */
#define MODE_VCO1_LCP_USB     5     /* Mc N03L2 LCP USB in VC01; hmm, was not 2-fanout, but 4... */
#define MODE_VCO2_USB         6     /* Mc F03L2 CH01: Mc VC02 USB (RCP at 1650.49MHz). */
                                    /* fanout_def = A : &CH01 : sign : 1:  2:  4:  6:  8; */
                                    /* fanout_def = A : &CH01 :  mag : 1: 10: 12: 14: 16; */
#define MODE_VCO1_CH2_LCP_USB 7     /* Mc F03L2 CH02: Mc VC01 USB (LCP at 1650.49MHz). */
                                    /* fanout_def = A : &CH02 : sign : 1: 18: 20: 22: 24; */
                                    /* fanout_def = A : &CH02 :  mag : 1: 26: 28: 30: 32; */

#define DATAMODE              MODE_2BIT_TEST
#define SAMPLES_IN_VSI_WORD   (16)              // depends on mode...

int
main(int argc, char **argv)
{
  int blksize;  /* memory / I/O block in traditional bytes */
  typedef unsigned long int tVsi;
  int blkwords;  /* mem / I/O in 32-bit words */
  int stripParity;
  int nonparityblkwords;  /* less parity words */
  int fftsize;
  int ffts_per_block;
  int totalblks;  /* how many blocks to read from file */
  char *memblk;  /* file I/O pointer */
  char *nextaftermemblk;  /* to check the whole block has been consumed */

  fftw_real *in, *out;  /* libfftw in and out arrays, sized by 'fftsize' */
  fftw_real *power_spectrum, *phase;  /* out-->pwr/phase, sized by fftsize/2+1 */
  rfftw_plan p;
  int fid;  /* opened input file number */
  int k;
  tVsi prevMemwrd = 0;  /* previous VSI mem word for NRZM-decoding */
  int decodeNRZM;

  if (argc < 7) {
    fprintf(stderr, "%s: needs at least blocksize stripParity decodeNRZM fftsize|0 totalblocks path\n", argv[0]);
    exit(EXIT_FAILURE);
  }

  /* File read block size and the corresponding memblk of 32-bit words. */
  blksize = atoi(argv[1]);
  if (blksize % sizeof(tVsi) != 0) {
    fprintf(stderr, "%s: blocksize has to be a multiple of %d\n", argv[0], sizeof(tVsi));
    exit(EXIT_FAILURE);
  }
  blkwords = blksize / sizeof(tVsi);

  assert( (memblk = malloc(blksize)) != NULL );
  nextaftermemblk = &(memblk[blksize]);

  /* Strip parity bits? */
  stripParity = atoi(argv[2]);
  if (stripParity) {
    nonparityblkwords = blkwords * 8 / 9;
    /* xxx: missing check that blocksize has multiples of 9... */
  } else {
    nonparityblkwords = blkwords;
  }

  /* Decode NRZM? */
  decodeNRZM = atoi(argv[3]);

  /* FFT bins. */
  fftsize = atoi(argv[4]);

  if (fftsize % SAMPLES_IN_VSI_WORD != 0) {
    fprintf(stderr, "%s: fftsize has to be a multiple of %d\n", argv[0], SAMPLES_IN_VSI_WORD);
    exit(EXIT_FAILURE);
  }
  if (fftsize < SAMPLES_IN_VSI_WORD) {
    /* Calculate default, the whole memblk to one FFT transform. */
    fftsize = nonparityblkwords * SAMPLES_IN_VSI_WORD;
    fprintf(stderr, "%s: defaulting fftsize to %d\n", argv[0], fftsize);
  }
  if (((nonparityblkwords * SAMPLES_IN_VSI_WORD) % fftsize) != 0) {
    fprintf(stderr, "%s: blocksize (in bits %d) has to be a multiple of fftsize (%d)\n", argv[0], (nonparityblkwords * SAMPLES_IN_VSI_WORD), fftsize);
    exit(EXIT_FAILURE);
  }
  ffts_per_block = (nonparityblkwords * SAMPLES_IN_VSI_WORD) / fftsize;
  assert( (in = (fftw_real *)calloc(fftsize, sizeof(fftw_real))) != NULL );
  assert( (out = (fftw_real *)calloc(fftsize, sizeof(fftw_real))) != NULL );
  assert( (power_spectrum = (fftw_real *)calloc(fftsize/2+1, sizeof(fftw_real))) != NULL );
  assert( (phase = (fftw_real *)calloc(fftsize/2+1, sizeof(fftw_real))) != NULL );

  totalblks = atoi(argv[5]);

  p = rfftw_create_plan(fftsize, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE);

  /* assert( (fid = open(argv[6], O_RDONLY)) != -1 ); */
  /* Try large file support. */
  assert( (fid = open64(argv[6], O_RDONLY)) != -1 );

  /* Start at position given in argv[7], 64 bit offset */
  if (argc >= 8) {
    assert((lseek(fid, atoll(argv[7]), SEEK_SET)) != -1);
  }

  fprintf(stderr, "totalblks = %d, ffts_per_block = %d\n", totalblks, ffts_per_block);
  // contents of in, out, power_spectrum, phase is already zeroed (calloc)

  /* Assemble "in" array from file - read all blocks, one by one, and convert data then sum data into "in". */
  for (k=0; k<totalblks; k++) {
    tVsi *memwords;  /* memory 32-bit word pointer (starts at memblk) */
    tVsi memwrd;  /* current NRZM-decoded VSI word */
    tVsi parwrd;  /* parity check word */
    int f, w;

    parwrd = 0xffffffff;

    /* Read one block to memory. */
    assert( (read(fid,
                  memblk,
                  blksize)) == blksize );

    memwords = (tVsi *)memblk;
    for (f=0; f<ffts_per_block; f++) {

      fftw_real *inp = in;
      int b;

      for (w=0; w<(fftsize/SAMPLES_IN_VSI_WORD); w++) {
        int twobits;
        const static double volts[4] = {
          /* 0=> */ -1.0,  /* <Vth, tape drive encoding Mk4 fmt */
          /* 1=> */ -0.5,  /* >=Vth..0 */
          /* 2=> */  0.5,  /* 0..<=Vth */
          /* 3=> */  1.0,  /* >Vth */
        };
        /* VLBA sampler encoding seems to be the same as on tape; */
        /* in Mk4 fmt sampler encoding was changed to simplify */
        /* comparator logic. */
        const static double vsvolts[4] = {
          /* 0=> */ -1.0,  /* <Vth, VLBA sampler encoding */
          /* 1=> */ -0.5,  /* >=Vth..0 */
          /* 2=> */  0.5,  /* 0..<=Vth(200mV) */
          /* 3=> */  1.0,  /* >Vth */
        };
        
        /* NRZM decoding using XOR with previous data word. */
        if (decodeNRZM) {
          memwrd = prevMemwrd ^ (*memwords);
        } else {
          memwrd = (*memwords);
        }

        /* Calculate parity word. */
        if (stripParity) {
          parwrd ^= memwrd;
        }

#if 0
        /* Debugging NRZM decoding. */
        fprintf(stdout, "w[%08X] = %08X\n", (unsigned int)w, (unsigned int) memwrd);
#endif

      // TODO: summing up the data into float variable array 'in' /may/ in certain 
      // cases lead to numeric overflow e.g. if the number of 'totalblocks' is really huge.

#if DATAMODE==MODE_2BIT_TEST
        /* VLBA sampler, VSIC mode 1, VSIB mode 7, "2-bit test". */
        /* (Has to swap sign/mag, just like with tape Mk4.) */
        { int i;
        const static double vsvolts2b[4] = { // in: 000ab   out: 00ba, remap of vsvolts[]
          /* 00: 0=> */ -1.0,  /* <Vth, VLBA sampler encoding */
          /* 01: 1=> */  0.5,  /* >=Vth..0 */
          /* 10: 2=> */ -0.5,  /* 0..<=Vth(200mV) */
          /* 11: 3=> */  1.0,  /* >Vth */
        };
        for (i=0; i<16; i++) {
          //twobits = (((memwrd & 0x00000001) << 1)&0x2)  /* b0, 1us+0 */
          //  | (((memwrd & 0x00000002) >> 1)&0x1);  /* b1, 1um+0 */
          //*inp++ = vsvolts[twobits];
          twobits = memwrd & (tVsi)0x03;
          *inp += vsvolts2b[twobits]; *inp++;
          memwrd >>= 2;
        }
        }
#elif DATAMODE==MODE_BBC_3LSB
        /* 3rd VC/BBC lower sideband, VC05 JB, BBC03 Mh. */
        twobits = ((memwrd & 0x00000002)     )&0x2  /* b1, 3ls+0 */
          || ((memwrd & 0x00000020) >> 5)&0x1;  /* b5, 3lm+0 */
        *inp += volts[twobits]; *inp++;
        twobits = ((memwrd & 0x00000008) >> 2)&0x2  /* b3, 3ls+1 */
          || ((memwrd & 0x00000080) >> 7)&0x1;  /* b7, 3lm+1 */
        *inp += volts[twobits]; *inp++;
#elif DATAMODE==MODE_16B_RAW_VLBA
        /* 16-bit raw VLBA sampler data, BBC01 USB in two LSBits of 16b.w. */
        /* (Has to swap sign/mag, just like with tape Mk4.) */
        twobits = (((memwrd & 0x00000001) << 1)&0x2)  /* b0, 1us+0 */
          | (((memwrd & 0x00000002) >> 1)&0x1);  /* b1, 1um+0 */
        *inp += vsvolts[twobits]; *inp++;
        twobits = (((memwrd & 0x00010000) >> 15)&0x2)  /* b16, 1us+1 */
          | (((memwrd & 0x00020000) >> 17)&0x1);  /* b17, 1um+1 */
        *inp += vsvolts[twobits]; *inp++;
#elif DATAMODE==MODE_BBC_2LSB
        /* 2nd VC/BBC lower sideband, VC03 JB, BBC02 Mh. */
        twobits = ((memwrd & 0x00000001) << 1)&0x2  /* b0, 1ls+0 */
          || ((memwrd & 0x00000010) >> 4)&0x1;  /* b4, 1lm+0 */
        *inp += volts[twobits]; *inp++;
        twobits = ((memwrd & 0x00000004) >> 1)&0x2  /* b2, 1ls+1 */
          || ((memwrd & 0x00000040) >> 6)&0x1;  /* b6, 1lm+1 */
        *inp += volts[twobits]; *inp++;
#elif DATAMODE==MODE_VCO1_LCP_USB
        /* Mc N03L2 LCP USB in VC01; hmm, was not 2-fanout, but 4... */
        twobits = ((memwrd & 0x01000000) >> 23)&0x2  /* b24, 1us+0 */
          || ((memwrd & 0x10000000) >> 28)&0x1;  /* b28, 1um+0 */
        *inp += volts[twobits]; *inp++;
        twobits = ((memwrd & 0x04000000) >> 25)&0x2  /* b26, 1us+1 */
          || ((memwrd & 0x40000000) >> 30)&0x1;  /* b30, 1um+1 */
        *inp += volts[twobits]; *inp++;
#elif DATAMODE=MODE_VCO2_USB
        /* Mc F03L2 CH01: Mc VC02 USB (RCP at 1650.49MHz). */
        /* fanout_def = A : &CH01 : sign : 1:  2:  4:  6:  8; */
        /* fanout_def = A : &CH01 :  mag : 1: 10: 12: 14: 16; */
        twobits = (memwrd <<  1)&0x2   /* b0, 2us+0 */
               || (memwrd >>  8)&0x1;  /* b8, 2um+0 */
        *inp += volts[twobits]; *inp++;
        twobits = (memwrd >>  1)&0x2   /* b2, 2us+1 */
               || (memwrd >> 10)&0x1;  /* b10, 2um+1 */
        *inp += volts[twobits]; *inp++;
        twobits = (memwrd >>  3)&0x2   /* b4, 2us+2 */
               || (memwrd >> 12)&0x1;  /* b12, 2um+2 */
        *inp += volts[twobits]; *inp++;
        twobits = (memwrd >>  5)&0x2   /* b6, 2us+3 */
               || (memwrd >> 14)&0x1;  /* b14, 2um+3 */
        *inp += volts[twobits]; *inp++;
#elif DATAMODE==MODE_VCO1_CH2_LCP_USB
        /* Mc F03L2 CH02: Mc VC01 USB (LCP at 1650.49MHz). */
        /* fanout_def = A : &CH02 : sign : 1: 18: 20: 22: 24; */
        /* fanout_def = A : &CH02 :  mag : 1: 26: 28: 30: 32; */
        twobits = (memwrd >> 15)&0x2   /* b16, 1us+0 */
               || (memwrd >> 24)&0x1;  /* b24, 1um+0 */
        *inp += volts[twobits]; *inp++;
        twobits = (memwrd >> 17)&0x2   /* b18, 1us+1 */
               || (memwrd >> 26)&0x1;  /* b26, 1um+1 */
        *inp += volts[twobits]; *inp++;
        twobits = (memwrd >> 19)&0x2   /* b20, 1us+2 */
               || (memwrd >> 28)&0x1;  /* b28, 1um+2 */
        *inp += volts[twobits]; *inp++;
        twobits = (memwrd >> 21)&0x2   /* b22, 1us+3 */
               || (memwrd >> 30)&0x1;  /* b30, 1um+3 */
        *inp += volts[twobits]; *inp++;
#endif
        prevMemwrd = *memwords;
        memwords++;

        if (stripParity) {
          /* Skip parity bit VSI words. */
          if ((w+1) % 8 == 0) {
            tVsi t;

            t = ((decodeNRZM)?(prevMemwrd ^ (*memwords)):(*memwords));
#if 0
            /* Debugging parity checking. */
            fprintf(stdout, "parwrd = %08X, t = %08X\n", (unsigned int)parwrd, (unsigned int)t);
#endif
            if ((parwrd == 0xffffffff) && (t == 0x0)) {
              /* sync word parity error, skip */
            } else if (parwrd != t) {
              fprintf(stderr, "%s: parity mismatch calc=%08X != %08X at block %d, fft %d, w=%d\n", argv[0], (unsigned int)parwrd, (unsigned int)t, k, f, w);
              exit(EXIT_FAILURE);
            }

            prevMemwrd = *memwords;
            memwords++;
            parwrd = 0xffffffff;
          }  /* if stripParity and 9th VSI word */
        }  /* if stripParity */
      }  /* for each word of one FFT */
    }  /* for every FFT transform in each memblk */

    /* After processing 'memblk' our pointer must point directly after it. */
    assert( (char *)memwords == nextaftermemblk );

  }  /* for every block in input file */

  assert( (close(fid) != -1) );

  /* FFT. */
  rfftw_one(p, in, out);
  power_spectrum[0] = out[0]*out[0];  /* DC component */
  /** phase[0] = 0.0; **/
  int b;
  for (b = 1; b < (fftsize+1)/2; ++b) {  /* (b < N/2 rounded up) */
     power_spectrum[b] += out[b]*out[b] + out[fftsize-b]*out[fftsize-b];
     /** phase[b] = atan2(out[fftsize-b], out[b])*360.0/M_PI; **/
  }
  if (fftsize % 2 == 0) {  /* fftsize is even */
     power_spectrum[fftsize/2] += out[fftsize/2]*out[fftsize/2];  /* Nyquist freq. */
     /** phase[fftsize/2] = 0.0; **/
  }
      
  /* Output spectrum, don't reverse freq scale even though LSB. */
  for (k=0; k<(fftsize/2); k++) {
    printf("%f\t%f\t%f\n", 32.0*(double)k/(double)fftsize, power_spectrum[k], phase[k]);
  }

  rfftw_destroy_plan(p);

  return (EXIT_SUCCESS);
}  /* main */
