/*************************************************
 *                                               *
 * Simple RAID to Network UDP data sender        *
 * (C) 2007 Jan Wagner                           *
 *                                               *
 * Connects to a client listening on the         *
 * specified port and sends the specified file   *
 * as UDP packets to the client.                 *
 * The transfer works like with UDP 'netcat',    *
 * with no protocol and no lost data retrans-    *
 * mission.                                      *
 *                                               *
 * Usage: raid2udp ip port packetsize ipd        *
 *                 file [filesize]               *
 *                                               *
 *   ip : client ip                              *
 *   port : client port                          *
 *   packetsize : size of the UDP packets        *
 *   ipd  : inter-packet delay in microseconds   *
 *   file : file to send                         *
 *   filesize : how many bytes to send           *
 *                                               *
 *************************************************/

//
// Compile
//   gcc -Wall raid2udp.c -o raid2udp -D_LARGEFILE64_SOURCE=1
//

#define _GNU_SOURCE 1
#define _LARGEFILE64_SOURCE 1  /* Large File Support (LFS) '*64()' functions. */
#define _FILE_OFFSET_BITS 64   /* Automatic '*()' --> '*64()' replacement. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <fcntl.h>
#include <unistd.h>

#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <arpa/inet.h>
#include <netdb.h>
#include <sys/socket.h>

#define EXIT_OK          0
#define EXIT_FILE_ERROR  1
#define EXIT_NET_ERROR   2

#define MIN(a,b) (a<b) ? a : b
typedef unsigned long long ull_t;

u_int64_t timedelta_us(struct timeval b, struct timeval e) {
   return ( 1e6*(e.tv_sec - b.tv_sec) - b.tv_usec + e.tv_usec );
}

void sleep_usec(u_int64_t usecs) {
   struct timeval start, now;
   gettimeofday(&start, NULL);
   do {
       gettimeofday(&now, NULL);
   } while (timedelta_us(start, now) < usecs);
}



int main(int argc, char* argv[]) {

  char     *address;
  int       udpport;
  int       packetsize;
  size_t    filesize;
  size_t    bytesremaining;
  size_t    packetcount;
  u_int64_t ipd;

  struct stat stbuf;
  char    *filename;
  int      file_fd;
  char*    mmbuf;
  char*    mmptr;

  int      tmp;

  struct sockaddr_in client_addr;
  int      socket_fd;

  if (argc < 6) {
    printf("Usage: raid2udp ip port packetsize ipd file [filesize]\n\n"
            "  ip : client ip\n"
            "  port : client port\n"
            "  packetsize : size of the UDP packets\n"
            "  ipd  : inter-packet delay in microseconds\n"
            "  file : file to send\n"
            "  filesize : how many bytes to send\n\n");
     return EXIT_OK;
  }


  /* get user params */
  address    = argv[1];
  udpport    = atoi(argv[2]);
  packetsize = atoi(argv[3]);
  ipd        = atol(argv[4]);
  filename   = argv[5];
  // file size later

  /* open the file */
  file_fd = open(filename, O_RDONLY);
  if (-1 == file_fd) {
     fprintf(stderr, "File open failed: '%s'\n", strerror(errno)); 
     return EXIT_FILE_ERROR;
  }  
  stat(filename, &stbuf);
  filesize = stbuf.st_size;
  if (argc == 7) {
     filesize = MIN(atol(argv[6]), filesize);
  }
  fprintf(stderr, "File opened, will send %Lu bytes.\n", (unsigned long long)filesize);

  /* map the file */
  mmbuf = (char*)mmap((caddr_t)0, filesize, PROT_READ, MAP_SHARED, file_fd, 0);
  if (MAP_FAILED == mmbuf) {
     fprintf(stderr, "File mmap failed: '%s'\n", strerror(errno)); 
     return EXIT_FILE_ERROR;
  }


  /* bind UDP socket */
  socket_fd = socket(AF_INET, SOCK_DGRAM, 0);
  if (socket_fd < 0) {
     fprintf(stderr, "Could not create a socket: '%s'\n", strerror(errno)); 
     return EXIT_NET_ERROR;
  }  
  tmp = 1;
  if (setsockopt(socket_fd, SOL_SOCKET, SO_NO_CHECK, &tmp, sizeof(tmp)) < 0) {
     fprintf(stderr, "Warning: disabling UDP checksums failed: '%s'\n", strerror(errno)); 
  }
  client_addr.sin_family      = AF_INET;
  client_addr.sin_port        = htons(udpport);
  client_addr.sin_addr.s_addr = inet_addr(address);


  /* prepare vars */
  mmptr                    = mmbuf; 
  bytesremaining           = filesize;
  packetcount              = 0;

  /* send all the data */
  fprintf(stderr, "Sending...\n");
  while (bytesremaining > 0) {
     ssize_t rc, wrlen;

     // send next packet
     wrlen = MIN(packetsize, bytesremaining);
     rc    = sendto(socket_fd, mmptr, wrlen, 0, (struct sockaddr*)&client_addr, sizeof(client_addr));
     if (rc < 0) {
        fprintf(stderr, "sendto() at byte %p failed: '%s'\n", 
                (void*)(mmptr-mmbuf), strerror(errno));
        break;
     }
     sleep_usec(ipd);

     // move forward
     bytesremaining -= rc;
     mmptr += rc;
     packetcount++;

     // interval report
     if (0 == (packetcount & 0x3FFF)) {
        fprintf(stderr, "  %6.3f%% done\n", 100.0 - (100.0*bytesremaining)/filesize);
     }
  }

  /* close the socket */
  close(socket_fd);

  /* release memory map */
  munmap((caddr_t)mmbuf, filesize);

  fprintf(stderr, " 100.000%% done.\nCompleted!\n");
  return EXIT_OK;
}
