/* 
 *  rtspectrum.c -- Semi-realtime Fourier transform for VSIB data. 
 *
 *  The tool combines wr.c and ft3.c, and reads the data directly from VSIB.
 *  Some data format parsing is currently hard-coded, but there is also the
 *  option to specify a custom format string from the command line.
 *
 *  Rtspectrum transforms data sets of size 'fftsize', and displays a power 
 *  spectrum with graphing "persistence" i.e. data points are FIR low-pass 
 *  filtered by 'fft_stacking_len'. An optional windowing and window overlap
 *  can be enabled to get slightly better data, though for overlapping the
 *  current producer/consumer implementation does not yet guarantee datablocks
 *  are consecutive.
 *
 *  If the incoming data rate is too fast compared to FFT/DFT throughput,
 *  then different skipping schemes may be adopted. Currently the scheme is
 *  to skip FFTs. Another scheme could be to occasionally reduce the number
 *  of FT bins and interpolate the coarser spectrum into the existing 
 *  earlier spectral points.
 *
 *  Requires: FFTW (fftw2-devel), pgplot (pgplot5), ncurses (ncurses5-devel)
 *
 *  Status: working
 *
 */


/* Copyright (C) 2002 Ari Mujunen, Ari.Mujunen@hut.fi
   Copyright (C) 2006 Jan Wagner, jan.wagner@iki.fi

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <signal.h>
#include <sys/time.h>

#include <curses.h>  /* for getch() */
#include <pthread.h> /* for threads */
#include <rfftw.h>
#include <fftw.h>
#include <math.h>  /* for M_PI, atan2() */

#include <unistd.h>  /* usleep() */

#include <sys/ipc.h>  /* for shared memory */
#include <sys/shm.h>

#include <sys/ioctl.h>
#include "vsib_ioctl.h"

#define VSIB_DEV_PATH    "/dev/vsib"  // hard-coded reference to VSIB device node
//#define VSIB_DEV_FROMCLI                // Define to read VSIB via STDIN (i.e. ./rtspectrum < /dev/vsib),
                                      // or comment out to use hard-coded VSIB_DEV_PATH

#include <cpgplot.h> /* for plotting */

#include <termios.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/types.h>
int my_getch(void);
int kbhit(void);


#define DEBUG 1


// FT3STYLE: define to use FT3.C style of abs(fft) + abs(fft) + abs(fft) + .. stacking
#define FT3STYLE yeah
//#define FFT_USE_WINDOWING yeah

// PGPLOT axis styles and scalings
#define PG_AXIS_LINEAR       0
#define PG_AXIS_XLOG         10
#define PG_AXIS_YLOG         20
#define PG_AXIS_XYLOG        30
//#define AXIS_STYLE           PG_AXIS_LINEAR // for linear y plot
#define AXIS_STYLE           PG_AXIS_YLOG   // for log y plot
#define PG_SCALE_INDEPENDENT 0
#define PG_SCALE_FIXED       1

#define GRAPHMODE_CONSOLE    0
#define GRAPHMODE_PGPLOT     1

#define PLOTSTYLE_SCATTER    0
#define PLOTSTYLE_LINES      1
#define PLOTSTYLE            PLOTSTYLE_LINES   // which plotting style to use

#define PLOT_REFRESHRATE_HZ  25 // limit for maximum plot refresh rate in Hz
#define PLOT_LAG_MULTIPLIER  32 // lag means how many VSIB blocks were read for every one 
                                // block processed by FFT, the multiplier specifies how many
                                // plottings to screen (but not FFT calculations) to skip before next plotting


//-----------------------------------------------------------------------------------
//  VSIB GLOBALS AND DEFINITIONS

/* (possibly outdated info modes.txt below:)
 
  VSIB Control Register Mode Bits
  -------------------------------
  alias cmode: STD_LOGIC_VECTOR(3 downto 0) is capture_mode(19 downto 16);
  all32bits:	"0000";	-- MK5 32-bit mode, all the VSI data
  even16bits:	"0001"; -- MK5 16-bit mode or all sign bits from VLBA sampler
  low16bits:	"0010"; -- alt: 2-bit USB from 8 BBCs
                        -- norm: 2-bit USB+LSB from 4 BBCs
                        -- also S2 16-bitstream mode
  low8bits:	"0011"; -- alt: 2-bit USB from 4 BBC:s
                        -- norm: 2-bit USB+LSB from 2 BBCs
  loweven8bits:	"0100"; -- MK5 8-bit mode or 8 sign bits from VLBA sampler
                        -- alt: 1-bit USB from 8 BBCs
                        -- norm: 1-bit USB+LSB from 4 BBCs
  autoc4x2bits:	"0101"; -- as 'low8bits'; Autocorrelator 2-bit mode
  autoc4x1bit:	"0110"; -- 4-bit low even; Autocorrelator 1-bit mode
  test2bits:	"0111"; -- 2-bit mode for test
  test1bit:	"1000"; -- 1-bit mode for test
  VLBA_142:	"1001"; -- VLBA 1:4 fanout mode with 2-bit sampling
  VLBA_141:	"1010"; -- VLBA 1:4 fanout mode with 1-bit sampling
  VLBA_122:	"1011"; -- VLBA 1:2 fanout mode with 2-bit sampling
  VLBA_121:	"1100"; -- VLBA 1:2 fanout mode with 1-bit sampling
  VLBA_112:	"1101"; -- VLBA 1:1 fanout mode with 2-bit sampling
  
*/

typedef unsigned long int vsi_t;        // VSI word typedef for 32-bit int (uint32)

enum vsibmode { all32bits=0, even16bits, low16bits, low8bits, loweven8bits, 
   autoc4x2bits, autoc4x1bit, test2bits, test1bit, VLBA_142, 
   VLBA_141, VLBA_122, VLBA_121, VLBA_112, // 13
   mode14, modeCustom // 15
} vsibmodes;

unsigned char samples_in_vsi_word_table[] = {
   16, 16, 16, 16, 16,   // vsib_mode 0..4
   16, 16, 16, 16, 16,   // vsib_mode 5..9
   16, 16, 16, 16, 16    // vsib_mode 10..14
};
unsigned char channels_in_vsi_word_table[] = {
   1, 1, 2, 16, 16,     // vsib_mode 0..4
   16, 16, 1, 16, 16,   // vsib_mode 5..9
   16, 16, 16, 16, 16   // vsib_mode 10..14
};

int vsib_fileno;                        // Should be STDIN_FILENO for reading per "< /dev/vsib"
int vsib_mode = 1;                      // Mode of VSIB board, see modes.txt
int vsib_mode_gigabit = 0;              // 0: gigabit mode off
int vsib_mode_embed_1pps_markers = 0;   // 0: don't embed 1pps markers
int vsib_mode_skip_samples = 0;         // 0: don't skip samples, 1: leave out every 2nd sample, 2: ...

const static double mk4volts[4] = {     // conversion from 2-bit sample value to "real" voltage level
   /* 0=> */ -1.0,  /* <Vth, tape drive encoding Mk4 fmt */
   /* 1=> */ -0.5,  /* >=Vth..0 */
   /* 2=> */  0.5,  /* 0..<=Vth */
   /* 3=> */  1.0,  /* >Vth */
};
const static double mk4voltsBR[4] = { // index bit-remapped version of mk4volts[] (in: 000ab   out: 00ba)
   /* 00: 0=> */ -1.0,  /* <Vth, VLBA sampler encoding */
   /* 01: 1=> */  0.5,  /* >=Vth..0 */
   /* 10: 2=> */ -0.5,  /* 0..<=Vth(200mV) */
   /* 11: 3=> */  1.0,  /* >Vth */
};

// Custom (nearly) arbitraty data format
// Format: "w[nn]b[xx],[yy],[zz],..." : width nn in bits, MSB is bit xx, next yy, LSB zz, ...
int userformat_bitwidth;                     // how many bits in one sample block
volatile vsi_t userformat_bit_mask[8*sizeof(vsi_t)];  // mask from which bits inside sample block to assemble the channel's value/sample
int userformat_bit_masklen;                  // length of mask (number of used values in mask)
vsi_t powOf2s[8*sizeof(vsi_t)];              // powers of 2 table to generate the above bitmask...


//-----------------------------------------------------------------------------------
//  FFT GLOBALS AND DEFINITIONS

// vsib unpacked data buffers
#define BUFFER_COUNT              8      // How many buffer blocks of size 'fftsize' to use
typedef unsigned char bufdata_t;         // Data type for unpacked buffer data, usually not floatingpoint
struct vsibBufferEntry_t {
   volatile unsigned char data_valid;    // 0: no or incomplete raw data, 1: valid raw data
   volatile unsigned char reservedByFFT; // 0: can be written to, 1: write blocked
   vsi_t             *rawdata;           // Where raw unprocessed VSIB data of size 'blksize'(rawdata_len) is placed
   int rawdata_len;                      // Length of rawdata in vsi_t words
   bufdata_t         *data;              // Where actual unpacked VSIB data of size 'fftsize'(data_len) is placed
   int data_len;                         // Length of data in bufdata_t words
} vsibBuffer[BUFFER_COUNT];              // Buffers for VSIB data input

unsigned int vsibBuffer_channel = 0;     // If VSI word contains multiple channels, which to extract
unsigned int vsibBuffer_writingAt = 0;    
unsigned int vsibBuffer_readingAt = 0; 

// FFTing throughput vs VSIB throughput lag handling
#define LAG_HM_FFTSKIPPING      0        // Skip FFTs
#define LAG_HM_FFTTHINNING      1        // Use smaller FFTs
#define LAG_HM_DFTINTERLEAVING  2        // Calculate sub spectrum DFTs and time-interleave
#define LAG_HM_DEFAULT          LAG_HM_FFTSKIPPING
long lag_counter = 0;                    // Counter for statistics and for adapting to too slow FFT,
                                         // increased by producer, checked by consumer (FFTing thread)
int lag_handling_mode = LAG_HM_DEFAULT;  // Use default lag handling mode initially...

// FFTing parameters from command line
unsigned int orig_samplerate;  // Original sampling rate MHz
double samplerate;             // Actual sampling rate after sample skipping
unsigned int fftsize;          // Number of points in FFT/DFT
unsigned int fft_stacking_len; // How many FFTs will be stacked (summed) before taking the abs() power spectrum

// FFTW data
fftw_real *fftw_in, *fftw_out;           // libfftw in and out arrays, sized by 'fftsize'
fftw_real *fftw_windowed_in;             // input array for windowed data from fftw_in
fftw_real *fftw_window;                  // precalculated hanning/hamming/... window weights
fftw_real *power_spectrum, *phase;       // out-->pwr/phase, sized by fftsize/2+1
fftw_real *stacked_complex_spectrum;     // fft_stacking_len number of summed up FFT's -- actually right now this is lowpass filtering, not stacking ;-)
rfftw_plan plan;
#if AXIS_STYLE==PG_AXIS_YLOG
fftw_real *power_spectrum_dB;
#endif

// DFT/FFT thread, consumer of VSIB data produced by the main() routine
void *ffting_thread(void* params);
volatile unsigned int ffting_thread_stop;

// raw data unpacking helper
void unpack_buffer_data(struct vsibBufferEntry_t *buf);



//-----------------------------------------------------------------------------------
//  KEY/SIGNAL FUNCTIONS

sig_atomic_t ctrlC_pressed = 0;

void SIGINT_handler(int signum)
{
   assert (signum == SIGINT);
   ctrlC_pressed = 1;
}

/*
 *  kbhit()  --  a keyboard lookahead monitor
 *
 *  returns the number of characters available to read.
 */
int kbhit(void)
{
  int             count = 0;
  int             error;
  struct timespec tv;
  struct termios  otty, ntty;

  tcgetattr(STDIN_FILENO, &otty);
  ntty = otty;
  ntty.c_lflag   &= ~ICANON; /* raw mode */

  if (0 == (error = tcsetattr(STDIN_FILENO, TCSANOW, &ntty))) {
    error        += ioctl(STDIN_FILENO, FIONREAD, &count);
    error        += tcsetattr(STDIN_FILENO, TCSANOW, &otty);
    /* minimal 10ms delay gives up cpu time slice,
     * allows use in a tight loop */
    tv.tv_sec     = 0;
    tv.tv_nsec    = 10;
    nanosleep(&tv, NULL);
  }

  return error == 0 ? count : -1;
} 

/*
 *  my_getch()  --  a blocking single character input from stdin
 *
 *  Returns a character, or -1 if an input error occurs.
 *
 */
int my_getch(void)
{
  unsigned char         ch;
  int                   error;
  struct termios        otty, ntty;

  fflush(stdout);
  tcgetattr(STDIN_FILENO, &otty);
  ntty = otty;

  ntty.c_lflag &= ~ICANON;   /* line settings    */
  ntty.c_lflag &= ~ECHO;     /* disable echo     */
  ntty.c_cc[VMIN]  = 1;      /* block for input  */
  ntty.c_cc[VTIME] = 0;      /* timer is ignored */

  if (0 == (error = tcsetattr(STDIN_FILENO, TCSANOW, &ntty))) {
    /* get a single character from stdin */
    error  = read(STDIN_FILENO, &ch, 1 );
    /* restore old settings */
    error += tcsetattr(STDIN_FILENO, TCSANOW, &otty);
  }

  return (error == 1 ? (int) ch : -1 );
} 


//-----------------------------------------------------------------------------------
//  VSIB FUNCTIONS


// vsib_ioctl(..) : A protected (error-checking) ioctl() for VSIB driver
static void vsib_ioctl(unsigned int mode, unsigned long arg) 
{
  if (ioctl(vsib_fileno, mode, arg)) {
    char err[255];
    snprintf(err, sizeof(err), "rtspectrum: ioctl(vsib_fileno=%d, mode=0x%04x,...)", vsib_fileno, mode);
    perror(err);
    fprintf(stderr, "rtspectrum: standard I/O is not an VSIB board\n");
    exit(EXIT_FAILURE);
  }
}

//
// start_VSIB() : initialize VSIB using settings already in the global variables
//
void start_VSIB(void)
{
  vsib_ioctl(VSIB_SET_MODE,
             (VSIB_MODE_MODE(vsib_mode)
              | VSIB_MODE_RUN
              | (vsib_mode_gigabit ? VSIB_MODE_GIGABIT : 0)
              | (vsib_mode_embed_1pps_markers ? VSIB_MODE_EMBED_1PPS_MARKERS : 0)
              | (vsib_mode_skip_samples & 0x0000ffff))
            );
}


/* Upcoming wrshm.h: */
#define fourCharLong(a,b,c,d) ( ((long)a)<<24 | ((long)b)<<16 | ((long)c)<<8 | d )
typedef struct sSh {
  int relSeekBlocks;
} tSh, *ptSh;




//-----------------------------------------------------------------------------------
//  MAIN

int main(int argc, char **argv)
{

  int blksize;                   // How many bytes(!) to read from VSIB as to 
                                 // unpack the data into one fftsize'd block
  int vsisize;                   // Same but in VSI words

  pthread_t consumerthread;      // Does the FFT and plot stuff
  struct sigaction sa;

  int usleeps;
  int k, rounds;

  /* Shared memory unique identifier (key) and the returned reference id. */
  key_t shKey;
  int shId = -1;
  ptSh sh;

  /* Info / usage */
  printf("\nRTSPECTRUM v1.0 -- a tool for vsib card quasi-realtime data monitoring\n");

  // TODO: check environment var PGPLOT_DIR=/usr/lib/pgplot is set?
        
  if (argc < 8) {
    fprintf(stderr, "\nUse:\n"
    #ifdef VSIB_DEV_FROMCLI
      " rtspectrum  samplerate fftsize fftstacklen channel mode skip giga [custom mode string cms] < /dev/vsib\n\n"
    #else
      " rtspectrum  samplerate fftsize fftstacklen channel mode skip giga [custom mode string cms]\n\n"
    #endif
      "About the parameters:\n"
      "   samplerate   : original sampling rate in MHz\n"
      "   fftsize      : maximum number of FFT/DFT bins attempted to use\n"
      "   fftstacklen  : how many FFTs to stack/filter for one power spectrum\n"
      "   channel      : which channel to plot, 0 is the first\n"
      "   mode         : VSIB board mode (0=32bit, 1=16bit VLBA, ...)\n"
      "   skip         : number of samples to skip (0..N),\n"
      "                  sampling rate will be corrected automatically\n"
      "   giga         : 0 to disable, 1 to enable gigabit mode\n"
      "If mode is 0 or 15 (==custom mode):\n"
      "   cms          : bit format string (note: channel is ignored)\n"
      "Custom format example: \"w16b15,12\" for 16-bit width MSB=bit 15 LSB= bit 12\n"
      "Notes: fftsize of a power of 2 is the fastest\n"
      "\n"
    #ifndef VSIB_DEV_FROMCLI
      "NOTE: rtspectrum in '/dev/vsib' hardcoded mode.\n"
      "Use keys 'a' for previous, 's' for next channel. 'q' quits.\n"
    #endif
    );
    exit(EXIT_FAILURE);
  }
    
  /* Register new SIGINT handler */
  sigemptyset (&sa.sa_mask);
  sa.sa_flags = 0;
  sa.sa_handler = SIGINT_handler;
  sigaction (SIGINT, &sa, 0);

  /* Process args */
  orig_samplerate = atoi(argv[1]);
  if (orig_samplerate<1 || orig_samplerate>256) {
     orig_samplerate = 16; 
     fprintf(stderr, "Defaulting to sampling rate %d (MHz)\n", orig_samplerate);
  }
  fftsize = atoi(argv[2]); // TODO: maybe zero-pad fftsize'd buffers to size 2^N
  if (fftsize<16) {
     fftsize = 2*orig_samplerate*1000; // approx. kHz resolution (also depends on vsib skip...)
     fprintf(stderr, "Defaulting to fftsize of %d\n", fftsize);
  } 
  fft_stacking_len = atoi(argv[3]);
  if (fft_stacking_len<1) {
     fft_stacking_len = 16;
  }
  vsibBuffer_channel = atoi(argv[4]);
  vsib_mode = atoi(argv[5]);
  vsib_mode_skip_samples = atoi(argv[6]);
  samplerate = orig_samplerate / (vsib_mode_skip_samples + 1);
  vsib_mode_embed_1pps_markers = 0;
  vsib_mode_gigabit = atoi(argv[7]);

  #ifdef VSIB_DEV_FROMCLI
  vsib_fileno = STDIN_FILENO;
  #else
  vsib_fileno = open(VSIB_DEV_PATH, O_RDONLY);
  #endif
  
  #ifndef VSIB_DEV_FROMCLI
  /* disable echoing keyboard input */
  {
  static struct termios otty, ntty; 
  tcgetattr(STDIN_FILENO, &otty);
  ntty = otty;
  ntty.c_lflag &= ~ECHO;
  tcsetattr(STDIN_FILENO, TCSANOW, &ntty); 
  }
  #endif
  
  /* If custom formatting string, prepare tables */
  if (vsib_mode==modeCustom || vsib_mode==0) {
     if (argc<9) {
        fprintf(stderr, "Error: mode was %d but required custom formatting string was not specified!\n\n", vsib_mode);
        exit(EXIT_FAILURE);
     }
     vsib_mode = 0; // VSIB mode 0: all 32 bits
     // first clear the tables
     userformat_bitwidth = 0;
     userformat_bit_masklen = 0;
     for (k=0; k<(8*sizeof(vsi_t)); k++) {
        userformat_bit_mask[k] = 0; // 0 as "EOF" of bit mapping stream
     }
     powOf2s[0] = 1;
     for (k=1; k<(8*sizeof(vsi_t)); k++) {
       powOf2s[k] = ((vsi_t)2) * powOf2s[k-1];
     }
     // format "check", then build tables from format string
     // format: w16b15,12  == width 16 bits, take bits 15 (MSB) and 12 (LSB)
     assert(NULL != strchr(argv[8], 'w'));
     assert('w' == argv[8][0]);
     assert(NULL != strchr(argv[8], 'b'));
     assert(strlen(argv[8])>=5);
     {
       char *str_running = strdup(&argv[8][1]); // skip the initial 'w'
       char *str_token;
       int  ibit;
       int  bitnr;
       str_token = strsep(&str_running, "b,");
       // get the bit width
       userformat_bitwidth = atoi(str_token);
       assert(userformat_bitwidth>0 && userformat_bitwidth<=32);
       fprintf(stderr, "custom format bit window width = %d\n", userformat_bitwidth);
       // get the bit numbers
       ibit = 0;
       while (NULL != (str_token = strsep(&str_running, "b,"))) {
          bitnr = atoi(str_token);
          assert(bitnr>=0 && bitnr<(8*sizeof(vsi_t)));
          userformat_bit_mask[ibit] = powOf2s[bitnr];
          fprintf(stderr, "  take bit %d from bit %d (mask %ld)\n", ibit, bitnr, userformat_bit_mask[ibit]);
          ibit++;
       }
       assert(ibit<=userformat_bitwidth);
       fprintf(stderr, "  parse done, ibit = %d\n", ibit);
       userformat_bit_masklen = ibit;
     }
   }
  
  /* Show summary of settings */  
  fprintf(stderr, "rtspectrum settings used: %d MHz basehttp://www.kurp.hut.fi/, %f MHz actual, %d-point DFT/FFT, channel %d, vsib mode %d, skip %d, giga %d\n",
     orig_samplerate, samplerate, fftsize, vsibBuffer_channel, vsib_mode, vsib_mode_skip_samples, vsib_mode_gigabit);

  /* Calc size of VSIB chunks to process - blksize is how many bytes are needed to get fftsize samples */
  if (vsib_mode == 0) {
     // (N / sampsPerVSIword) * bytesPerVSIword; sampsPerVSIword=userformat_bitwidth/32bit
     blksize = sizeof(vsi_t) * fftsize * 32/userformat_bitwidth;
  } else if (vsib_mode<sizeof(samples_in_vsi_word_table) && vsib_mode<sizeof(channels_in_vsi_word_table)) {
     blksize = sizeof(vsi_t) * fftsize / 
       ( samples_in_vsi_word_table[vsib_mode] / channels_in_vsi_word_table[vsib_mode] );
  } else {
     blksize = sizeof(vsi_t) * fftsize / 16; // assume: 16 samples per 32bit, each from same channel
  }
  vsisize = blksize / sizeof(vsi_t);
  fprintf(stderr, "calculated VSIB blocksize of %d bytes\n", blksize);
  fprintf(stderr, "wr equiv: ./wr %d n %d %d 0 %d 1 /file.bin n < /dev/vsib\n", blksize, vsib_mode, vsib_mode_skip_samples, vsib_mode_gigabit);
  
  /* Alloc the buffers */
  for (k=0; k<BUFFER_COUNT; k++) { 
     // unpacked data buffer
     vsibBuffer[k].data_len = fftsize;
     vsibBuffer[k].data = (bufdata_t*)malloc(fftsize*sizeof(bufdata_t));
     if (NULL == vsibBuffer[k].data) {
        fprintf(stderr, "Error: could not allocate memory for unpack buffer %d of %d\n", k+1, BUFFER_COUNT);
        exit(EXIT_FAILURE);
     }
     // raw data buffer
     vsibBuffer[k].rawdata_len = vsisize;
     vsibBuffer[k].rawdata = (vsi_t*)malloc(vsisize*sizeof(vsi_t));
     if (NULL == vsibBuffer[k].rawdata) {
        fprintf(stderr, "Error: could not allocate memory for raw buffer %d of %d\n", k+1, BUFFER_COUNT);
        exit(EXIT_FAILURE);
     }
     vsibBuffer[k].data_valid = 0;
     vsibBuffer[k].reservedByFFT = 0;
  }
  vsibBuffer_writingAt = 0;
  vsibBuffer_readingAt = 0;
  
  
  /* Create and initialize 'wr<-->control' shared memory. */
  shKey = fourCharLong('v','s','i','b');
  assert( (shId = shmget(shKey, sizeof(tSh), IPC_CREAT | 0777)) != -1 );
  assert( (sh = (ptSh)shmat(shId, NULL, 0)) != (void *)-1 );
  sh->relSeekBlocks = 0;


  /* Set up FFTW */
  assert( (fftw_in = (fftw_real *)calloc(fftsize, sizeof(fftw_real))) != NULL );
  assert( (fftw_windowed_in = (fftw_real *)calloc(2*fftsize, sizeof(fftw_real))) != NULL );
  assert( (fftw_window = (fftw_real *)calloc(fftsize, sizeof(fftw_real))) != NULL );  
  assert( (fftw_out = (fftw_real *)calloc(fftsize, sizeof(fftw_real))) != NULL );
  assert( (stacked_complex_spectrum = (fftw_real *)calloc(fftsize, sizeof(fftw_real))) != NULL );
  assert( (power_spectrum = (fftw_real *)calloc(fftsize/2+1, sizeof(fftw_real))) != NULL );
  #if AXIS_STYLE==PG_AXIS_YLOG
  assert( (power_spectrum_dB = (fftw_real *)calloc(fftsize/2+1, sizeof(fftw_real))) != NULL );
  #endif
  assert( (phase = (fftw_real *)calloc(fftsize/2+1, sizeof(fftw_real))) != NULL );  
  plan = rfftw_create_plan(fftsize, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE);
  lag_counter = 0;

  
  /* Start FFTing until user interrupts with keypress */
  
  fprintf(stderr, "--spectrum calc started, press Ctrl-C to interrupt--\n");
  
  // start VSIB board from global settings
  start_VSIB();
  
  // start FFTing thread
  ffting_thread_stop = 0;
  pthread_create(&consumerthread, NULL, ffting_thread, NULL);
  
  usleeps = 0;
  rounds = 0;
  
  //while (ERR != getch()) {
  //while (wgetch(stdscr) != KEY_ENTER) {
  //while(++rounds < 64) {
  while (0 == ctrlC_pressed) {

    int        l;

    size_t  nread = 0;
    vsi_t   *vsib_in;   // Points to current writing location for VSIB input data
    size_t  remaining;  // How many bytes remainig till full block

    /* Find the next writeable buffer */

    for (l=0; l<BUFFER_COUNT; l++) {
       ++vsibBuffer_writingAt;
       vsibBuffer_writingAt = vsibBuffer_writingAt % BUFFER_COUNT;
       if (vsibBuffer[vsibBuffer_writingAt].reservedByFFT == 0) {
          break;
       }
    }
    vsibBuffer[vsibBuffer_writingAt].data_valid = 0;    
    
    // pointer where to write and how many bytes
    vsib_in = vsibBuffer[vsibBuffer_writingAt].rawdata;
    remaining = vsibBuffer[vsibBuffer_writingAt].rawdata_len * sizeof(vsi_t);
    
    /* Read as many bytes from VSIB as requested */
    #ifdef DEBUG
    //fprintf(stderr, "main: trying to write %d bytes to buf nr %d\n", remaining, vsibBuffer_writingAt);
    #endif
    do {
      nread = read (vsib_fileno, (char*)vsib_in, remaining);
      remaining -= nread;
      vsib_in = (vsi_t*) (((char*)vsib_in) + nread);
      if (remaining>0) {
         // not one full block in buffer, wait a small amount, 
         // probably ends up to be 10--20msec
         usleep(1000);
         usleeps++;
      }
    } while (remaining>0 && !ctrlC_pressed);
    
    vsibBuffer[vsibBuffer_writingAt].data_valid = 1;
    #ifdef DEBUG
    //fprintf(stderr, "main: new vsib data written to buf %d\n", vsibBuffer_writingAt);
    #endif

    // increase the lag counter (the FFTing thread will check this value to
    // see how many data blocks have been input and possibly lost while
    // doing FFT in the background)
    lag_counter++;

    // check and process keyboard input (only if not </dev/vsib redirection to STDIN...)
    #ifndef VSIB_DEV_FROMCLI
    while(kbhit()>0) {
      int kk, ok; vsi_t pow2val;
      int kbdChar = my_getch();
      if(-1 == kbdChar) { continue; }

      switch(kbdChar) {
         case 'a':
         case 'A':
            // previous channel
            // special adjustments for custom format mode:
            if (0==vsib_mode || 15==vsib_mode) {
               ok = 1;
               pow2val = powOf2s[userformat_bit_masklen];
               for (kk=0; kk<userformat_bit_masklen; kk++) {
                  if ((userformat_bit_mask[kk]/pow2val) < 1) { ok = 0; break; }
               }
               if (1==ok) {
                 for (kk=0; kk<userformat_bit_masklen; kk++) {
                     userformat_bit_mask[kk] /= pow2val;
                  }
                  vsibBuffer_channel--;
                  fprintf(stderr, "channel down: div bitmask by %ld\n", pow2val);
               } else {
                  fprintf(stderr, "channel down: already reached LSB bit in VSI Word\n");
               }
            } else {
               if (vsibBuffer_channel>0) vsibBuffer_channel--;
            }
            break;
         case 's':
         case 'S':
            // next channel
            // special adjustments for custom format mode:
            if (0==vsib_mode || 15==vsib_mode) {
               ok = 1;
               pow2val = powOf2s[userformat_bit_masklen];
               for (kk=0; kk<userformat_bit_masklen; kk++) {
                  if (userformat_bit_mask[kk] >= powOf2s[userformat_bitwidth-userformat_bit_masklen]) { ok = 0; break; }
               }
               if (1==ok) {
                  for (kk=0; kk<userformat_bit_masklen; kk++) {
                     userformat_bit_mask[kk] *= pow2val;
                  }
                  vsibBuffer_channel++;
                  fprintf(stderr, "channel up: mul bitmask by %ld\n", pow2val);
               } else {
                  fprintf(stderr, "channel up: already reached MSB bit in VSI Word\n");
               }
            } else {
               if(vsibBuffer_channel<16) vsibBuffer_channel++;
            }
            break;
         case 'q':
         case 'Q':
            ctrlC_pressed = 1;
            break;
         default: 
            fprintf(stderr, "Keys: 'a' for previous channel, 's' for next, 'q' to quit.\n");
            if (vsib_mode!=0 && vsib_mode!=15) {
               fprintf(stderr, "NOTE: vsib_mode is %d, 'a' 's' currently work properly only for mode 0!\n", vsib_mode);
            }
            break;
      }      
    }
    #endif

  } /* while() no Ctrl-C pressed */
     
  /* End the FFTing thread */
  fprintf(stderr, "Waiting for FFT processing to stop...\n");
  ffting_thread_stop = 1; // poor way to signal the thread to stop...
  pthread_join(consumerthread, NULL);
  
  /* Finalize FFTW plan */
  rfftw_destroy_plan(plan);

  /* Stop the board, first DMA, and when the last descriptor */
  /* has been transferred, then write stop to board command register. */
  vsib_ioctl(VSIB_DELAYED_STOP_DMA, 0);
  {
    unsigned long b;

    vsib_ioctl(VSIB_IS_DMA_DONE, (unsigned long)&b);
    while (!b) {
      fprintf(stderr, "Waiting for last DMA descriptor (sl=%d)\n",
              usleeps);
      usleep(100000);
      usleeps++;
      vsib_ioctl(VSIB_IS_DMA_DONE, (unsigned long)&b);
    }
  }
  vsib_ioctl(VSIB_SET_MODE, VSIB_MODE_STOP);

  /* Remove shared memory to mark that 'wr/rd' is no more running. */
  if ((shId != -1) && (sh != (ptSh)-1) && (sh != NULL)) {
    assert( shmctl(shId, IPC_RMID, NULL) == 0 );
    assert( shmdt(sh) == 0 );
  }  // if shared memory was allocated
  
  #ifndef VSIB_DEV_FROMCLI
  close(vsib_fileno);
  #endif

  #ifndef VSIB_DEV_FROMCLI
  /* enable echoing keyboard input again */
  {
  static struct termios otty, ntty; 
  tcgetattr(STDIN_FILENO, &otty);
  ntty = otty;
  ntty.c_lflag |= ECHO;
  tcsetattr(STDIN_FILENO, TCSANOW, &ntty); 
  }
  #endif
  
  fprintf(stderr, "Done!\n");
  return (EXIT_SUCCESS);
}  /* main */




//-----------------------------------------------------------------------------------
//  RAW VSIB DATA BIT UNPACKING


void unpack_buffer_data(struct vsibBufferEntry_t *buf)
{
    if (NULL == buf) { return; }
    
    /* Unpack the bit-packed samples of the desired channel into same buf */
    
    bufdata_t  *unpacked_out  = buf->data;
    int        max_out_len    = buf->data_len; // == fftsize
    int        out_len        = 0;

    vsi_t      vsi_word       = 0;
    int        l;
    unsigned char bit = 0;

    if ((NULL == buf->rawdata || buf->rawdata_len <= 0)
        || (NULL == unpacked_out || max_out_len <= 0)) { return; }
    
    //    buf->rawdata : 32bit input data of length buf->rawdata_len in vsi_t words
    //    unpacked_out : target bufdata_t* to unpacked data of 'max_out_len' length in bufdata_t words

    for (l=0; l<(buf->rawdata_len) && (out_len<max_out_len); l++) {

       //int twobits;                    // temp storage for bits of a single sample
       vsi_word = *(buf->rawdata + l);   // the current vsi word to process

       switch(vsib_mode) {
          case all32bits:    // 0: 32-bit mode
          case modeCustom:   // 15: dummy mode
             // Assumption: mode 0 wasn't set from command line, instead mode 15 "custom data format"
             // was specified, together with a bit format string...
             { 
             int i=0, j=0;
             for (i=0; i<(32/userformat_bitwidth); i++) {
                *unpacked_out = 0;
                for (j=0; j<userformat_bit_masklen; j++) {
                  bit = 0 != (vsi_word & userformat_bit_mask[j]);
                  *unpacked_out = (*unpacked_out << 1) | bit;
                }
                ++unpacked_out;
                ++out_len;
                vsi_word = vsi_word << userformat_bitwidth;
             }
             }
             break;
          case 1:
          case 7:
             /* VLBA sampler, VSIC mode 1, VSIB mode 7, "2-bit test". */
             /* Only one channel with two bits per sample */
             { 
             int i;
             for (i=0; i<16; i++) {
                //*unpacked_out++ = (bufdata_t) (vsi_word & (vsi_t)0x03); // 2 bits, no bit reversion here...
                // TODO: channel select...
                *unpacked_out++ = (bufdata_t) ((vsi_word & (vsi_t)0x01)<<1 | (vsi_word & (vsi_t)0x02)>>1); // 2 bits, bits reversed
                vsi_word >>= 2;
                ++out_len;
             }
             }
             break;
          case 2:
             /* 3rd VC/BBC lower sideband, VC05 JB, BBC03 Mh. */
             /* For volts conversion use mk4volts[] */
             if (0==vsibBuffer_channel) {
               *unpacked_out++ = (bufdata_t) ( ((vsi_word & 0x00000002)&0x2)  /* b1, 3ls+0 */
                 | (((vsi_word & 0x00000020) >> 5)&0x1) );  /* b5, 3lm+0 */
             } else {
               *unpacked_out++ = (bufdata_t) ( (((vsi_word & 0x00000008) >> 2)&0x2)  /* b3, 3ls+1 */
                 | (((vsi_word & 0x00000080) >> 7)&0x1) );  /* b7, 3lm+1 */
             }
             ++out_len;
             break;
          default:
             break;
        }
 //~ #elif DATAMODE==MODE_16B_RAW_VLBA
        //~ /* 16-bit raw VLBA sampler data, BBC01 USB in two LSBits of 16b.w. */
        //~ /* (Has to swap sign/mag, just like with tape Mk4.) */
        //~ twobits = (((memwrd & 0x00000001) << 1)&0x2)  /* b0, 1us+0 */
          //~ | (((memwrd & 0x00000002) >> 1)&0x1);  /* b1, 1um+0 */
        //~ *inp++ = vsvolts[twobits];
        //~ twobits = (((memwrd & 0x00010000) >> 15)&0x2)  /* b16, 1us+1 */
          //~ | (((memwrd & 0x00020000) >> 17)&0x1);  /* b17, 1um+1 */
        //~ *inp++ = vsvolts[twobits];
//~ #elif DATAMODE==MODE_BBC_2LSB
        //~ /* 2nd VC/BBC lower sideband, VC03 JB, BBC02 Mh. */
        //~ twobits = ((memwrd & 0x00000001) << 1)&0x2  /* b0, 1ls+0 */
          //~ || ((memwrd & 0x00000010) >> 4)&0x1;  /* b4, 1lm+0 */
        //~ *inp++ = volts[twobits];
        //~ twobits = ((memwrd & 0x00000004) >> 1)&0x2  /* b2, 1ls+1 */
          //~ || ((memwrd & 0x00000040) >> 6)&0x1;  /* b6, 1lm+1 */
        //~ *inp++ = volts[twobits];
//~ #elif DATAMODE==MODE_VCO1_LCP_USB
        //~ /* Mc N03L2 LCP USB in VC01; hmm, was not 2-fanout, but 4... */
        //~ twobits = ((memwrd & 0x01000000) >> 23)&0x2  /* b24, 1us+0 */
          //~ || ((memwrd & 0x10000000) >> 28)&0x1;  /* b28, 1um+0 */
        //~ *inp++ = volts[twobits];
        //~ twobits = ((memwrd & 0x04000000) >> 25)&0x2  /* b26, 1us+1 */
          //~ || ((memwrd & 0x40000000) >> 30)&0x1;  /* b30, 1um+1 */
        //~ *inp++ = volts[twobits];
//~ #elif DATAMODE=MODE_VCO2_USB
        //~ /* Mc F03L2 CH01: Mc VC02 USB (RCP at 1650.49MHz). */
        //~ /* fanout_def = A : &CH01 : sign : 1:  2:  4:  6:  8; */
        //~ /* fanout_def = A : &CH01 :  mag : 1: 10: 12: 14: 16; */
        //~ twobits = (memwrd <<  1)&0x2   /* b0, 2us+0 */
               //~ || (memwrd >>  8)&0x1;  /* b8, 2um+0 */
        //~ *inp++ = volts[twobits];
        //~ twobits = (memwrd >>  1)&0x2   /* b2, 2us+1 */
               //~ || (memwrd >> 10)&0x1;  /* b10, 2um+1 */
        //~ *inp++ = volts[twobits];
        //~ twobits = (memwrd >>  3)&0x2   /* b4, 2us+2 */
               //~ || (memwrd >> 12)&0x1;  /* b12, 2um+2 */
        //~ *inp++ = volts[twobits];
        //~ twobits = (memwrd >>  5)&0x2   /* b6, 2us+3 */
               //~ || (memwrd >> 14)&0x1;  /* b14, 2um+3 */
        //~ *inp++ = volts[twobits];
//~ #elif DATAMODE==MODE_VCO1_CH2_LCP_USB
        //~ /* Mc F03L2 CH02: Mc VC01 USB (LCP at 1650.49MHz). */
        //~ /* fanout_def = A : &CH02 : sign : 1: 18: 20: 22: 24; */
        //~ /* fanout_def = A : &CH02 :  mag : 1: 26: 28: 30: 32; */
        //~ twobits = (memwrd >> 15)&0x2   /* b16, 1us+0 */
               //~ || (memwrd >> 24)&0x1;  /* b24, 1um+0 */
        //~ *inp++ = volts[twobits];
        //~ twobits = (memwrd >> 17)&0x2   /* b18, 1us+1 */
               //~ || (memwrd >> 26)&0x1;  /* b26, 1um+1 */
        //~ *inp++ = volts[twobits];
        //~ twobits = (memwrd >> 19)&0x2   /* b20, 1us+2 */
               //~ || (memwrd >> 28)&0x1;  /* b28, 1um+2 */
        //~ *inp++ = volts[twobits];
        //~ twobits = (memwrd >> 21)&0x2   /* b22, 1us+3 */
               //~ || (memwrd >> 30)&0x1;  /* b30, 1um+3 */
        //~ *inp++ = volts[twobits];
//~ #endif

    }  /* bit unpacking for 'vsisize' number of vsi_t data words */

    return;
}

     
     
//-----------------------------------------------------------------------------------
//  FOURIER TRANSFORM WORKER THREAD

void *ffting_thread(void* params) {

   int l, dlen, peak_idx;
   bufdata_t *unpacked_in;
   fftw_real value;
   fftw_real stackingW1, stackingW2;
   float *x_scale;
   float *y_tmp;
   float ylog_min, ylog_max, log10val;
   float x_min, x_max;
   char plottitle[256];

   struct timeval timev;
   double timeoflastplot = 0, timecurr = 0;

   int graphMode = GRAPHMODE_CONSOLE;

   int lag_adjust = 0, lag_adjust_curr = 0;

   /* Attempt to start PGPLOT and set up the graph */
   if (cpgopen("/XSERVE") >= 1) {
     graphMode = GRAPHMODE_PGPLOT;
     x_scale = (float*)malloc(sizeof(float)*fftsize);
     y_tmp = (float*)malloc(sizeof(float)*fftsize);
     for(l=0;l<fftsize;l++) { 
        *(x_scale+l) = samplerate * (float)l/((float)fftsize); 
     }
     x_min = 0.0;
     x_max = samplerate / 2.0;
   }
   cpgask(0); // disable "Type <RETURN> for next page:" prompting

   /* Precalculate some stuff */
   stackingW1 = ((fftw_real)(fft_stacking_len-1)/(fftw_real)fft_stacking_len);
   stackingW2 = 1.0 / (fftw_real)fft_stacking_len;
   for (l=0; l<fftsize; l++) { 
      // (note: fftsize here, vs later vsibBuffer[vsibBuffer_readingAt].data_len - these have the same value, for now...)
      // Hanning:
      //fftw_window[l] = 0.5 + 0.5 * cos(3.14159265 * (double)(l-(fftsize/2)) / (double)(fftsize/2));
      // Hamming:
      fftw_window[l] = 0.54 - 0.46 * cos(2 * 3.14159265 * l / (double)(fftsize));
   }
   
   /* Loop for FFTs */
   
   while(0 == ffting_thread_stop) {

      /* Try to find next readable buffer */
      for (l=0; l<BUFFER_COUNT; l++) { 
        ++vsibBuffer_readingAt;
        vsibBuffer_readingAt = vsibBuffer_readingAt % BUFFER_COUNT;
        if (vsibBuffer[vsibBuffer_readingAt].data_valid == 1) {
           vsibBuffer[vsibBuffer_readingAt].reservedByFFT = 1; // reserve now
           break;
        }
      }
      // if no buffer found, try again later
      if (vsibBuffer[vsibBuffer_readingAt].data_valid == 0) {
        usleep(100); // maybe too long? system usleep reso is usually 10ms...
        continue; 
      } 
      
      // get to the data
      unpacked_in = (bufdata_t*)vsibBuffer[vsibBuffer_readingAt].data;
      dlen = vsibBuffer[vsibBuffer_readingAt].data_len;
            
      /* Unpack the raw data bits into unpacked buffer */
      unpack_buffer_data(&vsibBuffer[vsibBuffer_readingAt]);
            
      /* Convert bufdata_t into fftw_real using voltage tables */
      for (l=0; l<dlen; l++) {
        switch(vsib_mode) {
          case 0:
          case 15:
             if (userformat_bit_masklen>2) {
               value = (float) *(unpacked_in++);
             } else {
               value = mk4volts[*(unpacked_in++)];
             }
             break;
          case 1:
          case 2:
          case 7:
             //value = mk4voltsBR[*(unpacked_in++)];
             /*if (*unpacked_in>3) { 
                fprintf(stderr, "ffting_thread: illegal val %d at %d\n", *unpacked_in, l); 
                break; 
             }*/
             value = mk4volts[*(unpacked_in++)];
             break;
          default: // TODO
             break;
        } 
        *(fftw_in+l) = value;
      }
      
      /* Free the buffer so producer can write again, also remove 'valid' flag so we don't FFT it again */
      
      vsibBuffer[vsibBuffer_readingAt].reservedByFFT = 0;
      vsibBuffer[vsibBuffer_readingAt].data_valid = 0;
           
      /* Perform the DFT/FFT */
      #ifdef FFT_USE_WINDOWING
      // -- Perform with overlapping and windowing, thus FFT needs to be computed twice
      for (k=0; k<2; k++) {               
         // move older data to left
         // memcpy(to, from, bytecount);
         memmove((char*)&fftw_windowed_in[0], (char*)&fftw_windowed_in[dlen/2 - 1], sizeof(fftw_real)*dlen/2);
         // insert first (k=0) or second (k=1) half of new data at the right
         memcpy((char*)&fftw_windowed_in[dlen/2 - 1], (char*)&fftw_in[k*(dlen/2-1)], sizeof(fftw_real)*dlen/2);
         // copy the entire new block into right side of fftw_windowed_in and apply window to it
         memcpy((char*)&fftw_windowed_in[dlen], (char*)&fftw_windowed_in[0], sizeof(fftw_real)*dlen);
         for (l=0; l<dlen; l++) { fftw_windowed_in[dlen+l] *= fftw_window[l]; }
         // finally, FFT the windowed data
         rfftw_one(plan, (fftw_real*)&fftw_windowed_in[dlen], fftw_out);
      #else
         // -- Perform without overlap or windowing, may lead to artefacting/spurious
         rfftw_one(plan, fftw_in, fftw_out);
      #endif
         
         // TODO::  !! Proper stacking of FFTs and not just lowpass filtering !!
         //         !! Use the lag_counter for weighting the new FFT into stacked result i.e. estimate skipped FFTs !!
   
         #ifdef FT3STYLE
         
         // Ft3.c style : calc FFTs of fftsize ('dlen'), for every FFT power_spectrum+=abs(fftw_out), final sum is the spectrum
         *stacked_complex_spectrum = (*fftw_out) * (*fftw_out); // DC
         for (l=1; l<(dlen+1)/2; l++) {
           value = (*(fftw_out+l))*(*(fftw_out+l)) + (*(fftw_out+dlen-l))*(*(fftw_out+dlen-l)); // abs()
           *(stacked_complex_spectrum+l) = stackingW1*(*(stacked_complex_spectrum+l)) + stackingW2*value; // sum
           *(power_spectrum+l) = *(stacked_complex_spectrum+l);
         }
         if (dlen%2 == 0) { // DC overlapping Nyqyist?
           *(stacked_complex_spectrum+dlen/2) = (*(fftw_out+dlen/2)) * (*(fftw_out+dlen/2));
         }
         *power_spectrum = *stacked_complex_spectrum;
         *(power_spectrum+dlen/2) = *(stacked_complex_spectrum+dlen/2);
         
         #else
         
         // Complex spectrum stacking style
         for (l=0; l<dlen; l++) {
           *(stacked_complex_spectrum+l) = stackingW1*(*(stacked_complex_spectrum+l)) + stackingW2*(*(fftw_out+l));
         }
        
         // use stacked_complex_spectrum or fftw_out
         power_spectrum[0] = stacked_complex_spectrum[0]*stacked_complex_spectrum[0];  /* DC component */
         /** phase[0] = 0.0; **/
         for (l = 1; l < (dlen+1)/2; ++l) {  /* (b < N/2 rounded up) */
            power_spectrum[l] += stacked_complex_spectrum[l]*stacked_complex_spectrum[l] + stacked_complex_spectrum[dlen-l]*stacked_complex_spectrum[dlen-l];
            /** phase[b] = atan2(stacked_complex_spectrum[dlen-b], stacked_complex_spectrum[b])*360.0/M_PI; **/
         }
         if (dlen % 2 == 0) {  /* fftsize is even, Nyquist and DC overlap */
            power_spectrum[dlen/2] = stacked_complex_spectrum[dlen/2]*stacked_complex_spectrum[dlen/2];  /* Nyquist freq. */
            /** phase[dlen/2] = 0.0; **/
         }
         #endif
         
      #ifdef FFT_USE_WINDOWING
      }//for(fft overlapping)
      #endif
      
      #if AXIS_STYLE==PG_AXIS_YLOG
      power_spectrum_dB[0] = log10(power_spectrum[0]);
      #endif

      /* Skip plotting to screen if FFTing is too slow */
      if(lag_adjust_curr<lag_adjust) {
         
         // skip plots if too slow
         lag_adjust_curr++;
                  
      } else {
      
         /* Skip plots if updating too fast, limit is set by PLOT_REFRESHRATE_HZ */
         if(0 == gettimeofday(&timev, NULL)) {
             timecurr = (double)timev.tv_sec + (double)timev.tv_usec/1000000.0;
             if ((timecurr - timeoflastplot) < (1.0/PLOT_REFRESHRATE_HZ)) { continue; }
             timeoflastplot = timecurr;
         }

         // do a plot
         lag_adjust_curr = 0;

         /* Re-normalize and output spectrum, don't mirror x axis even if LSB. */
         value = 0.0;
         peak_idx = 0;
         ylog_min = 120.0; ylog_max = -120.0;
         for (l=1; l<(dlen/2); l++) { // start from 1 = exclude DC peaks
             if (power_spectrum[l]>value) {
                value = power_spectrum[l];
                peak_idx = l;
             }
             #if AXIS_STYLE==PG_AXIS_YLOG
             log10val = log10(power_spectrum[l]);
             power_spectrum_dB[l] = log10val;
             if (log10val>ylog_max) { ylog_max = log10val; }
             if (log10val<ylog_min) { ylog_min = log10val; }
             #endif
         }
         if (value > 1.0) {
           for (l=0; l<(dlen/2); l++) {
               power_spectrum[l] /= value;
               #if AXIS_STYLE==PG_AXIS_YLOG
               power_spectrum_dB[l] -= log10(value);
               #endif
           }
           #if AXIS_STYLE==PG_AXIS_YLOG
           ylog_max -= log10(value);
           ylog_min -= log10(value);
           #endif
         }
   
         switch(graphMode) {
         case GRAPHMODE_CONSOLE:
           printf("\033[2J"); // clear terminal
           printf("----------------------------------------------------\r\n");
           for (l=0; l<(dlen/2); l++) {
             int k;
             for(k=0;k<25*(power_spectrum[l]/value);k++) {
               putchar('#');
             }
             putchar('\r'); 
             putchar('\n'); 
           }
           printf("----------------------------------------------------\r\n");
           break;
         case GRAPHMODE_PGPLOT:
           //TODO: don't clear and repaint the entire "page" every time, use blitting or whatever for graph part
           //      and try to reduce flicker
           //cpgpage(); 
           cpgbbuf();
           #if AXIS_STYLE==PG_AXIS_LINEAR
           cpgenv(x_min, x_max, 0.0, 1.05, PG_SCALE_INDEPENDENT, PG_AXIS_LINEAR);
           sprintf(plottitle, "Normalized Spectrum : peak %5.2f @ %5.2f MHz, lag %ld, ch %d", 
                   value, x_scale[peak_idx], lag_counter, vsibBuffer_channel);
           cpglab("MHz", "FFT", plottitle);
           for (l=0; l<(dlen/2); l++) { 
              *(y_tmp+l) = (float)(power_spectrum[l]);  // convert, as PGPLOT wants float's...
           } 
           #elif AXIS_STYLE==PG_AXIS_YLOG
           cpgenv(x_min, x_max, ylog_min, ylog_max+0.05, PG_SCALE_INDEPENDENT, PG_AXIS_LINEAR);
           sprintf(plottitle, "Normalized dB Spectrum : peak %4.2f fake-dB @ %4.2f MHz, lag %ld, ch %d", 
                   log10(value), x_scale[peak_idx], lag_counter, vsibBuffer_channel);
           cpglab("MHz", "log10 FFT", plottitle);
           for (l=0; l<(dlen/2); l++) { 
              //*(y_tmp+l) = (float)(log10(power_spectrum[l]));  // convert to dB and float
              *(y_tmp+l) = (float)(power_spectrum_dB[l]);  // convert to float
           }          
           #else
              assert(false);
           #endif
           #if PLOTSTYLE==PLOTSTYLE_LINES
           // interpolated "line" plot:
           cpgline(dlen/2, x_scale, y_tmp);
           #else
           // dot "scatter" plot:
           cpgpt(dlen/2, x_scale, y_tmp, 1);
           #endif
           cpgebuf();
           break;
         default:
           break;
         }
      } // if(lag_adjust_curr>=lag_adjust)
      
      /* Adjust timing statistics */
      
      // TODO: really keep track of how many cycles of fftsize(/dlen) blocks
      // have been lost/skipped due to too slow FFT, then try
      // to adjust FFT processing somehow to speed up if desired
      lag_adjust = PLOT_LAG_MULTIPLIER*lag_counter;
      lag_counter = 0;
      
   } // while(not interrupted)

      
   /* Close PGPLOT */
   if(GRAPHMODE_PGPLOT==graphMode) { cpgclos(); } 
   
   return NULL;
   //pthread_exit(0);
}

/* 
 * $Log: rtspectrum.c,v $
 * Revision 1.21  2006/11/23 10:25:08  jwagner
 * mentioned ncurses package requirement
 *
 * Revision 1.20  2006/11/23 10:20:55  jwagner
 * fixed compiler warnings and read() bug
 *
 * Revision 1.19  2006/11/09 17:35:23  jwagner
 * updated comments
 *
 * Revision 1.18  2006/11/01 09:21:04  jwagner
 * reduced lag multiplier, Hz refresh rate seems more effective
 *
 * Revision 1.17  2006/10/10 11:37:40  jwagner
 * limit plot update rate for too fast plots, default max is 25 Hz
 *
 * Revision 1.16  2006/10/06 09:02:48  jwagner
 * user format channel up/down fix for non-32bit bitwidths
 *
 * Revision 1.15  2006/10/06 08:54:10  jwagner
 * added hard-coding /dev/vsib file as compile flag so not via STDIN, added channel select via keyboard e.g. STDIN
 *
 * Revision 1.14  2006/10/03 14:19:58  jwagner
 * added optional FFT overlap with windowing (Hann, Hamm)
 *
 * Revision 1.13  2006/10/03 12:53:09  jwagner
 * peak location now shown correctly
 * 
 * Revision 1.11  2006/10/03 12:43:16  jwagner
 * corrected x-axis labels and titles, corrected dB-spectr normalization and plot, removed duplicate log10() of spectrum calc, added CVS log
 *
 */
