/* sbbc.c -- Software Baseband Converter */


/* Copyright (C) 2006 Jan Wagner, jan.wagner@iki.fi

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <termios.h>
#include <signal.h>


#include <sys/ioctl.h>
#include "vsib_ioctl.h"
#define VSIB_DEV_PATH    "/dev/vsib"  // hard-coded reference to VSIB device node

typedef unsigned long int vsi_t;

//-----------------------------------------------------------------------------------
//  FILTER AND MIXER BLOCKS

#define BITS_PER_SAMP 2
double sampling_rate = 32e6;
struct _mixer_blk {
    float sinTab[1<<BITS_PER_SAMP + 1]; // 2-bit; 0 1.0 0 -1.0
    float cosTab[1<<BITS_PER_SAMP + 1]; // 2-bit: 1.0 0 -1.0 0
};

struct _filter_blk {
    int degree;
    float *coeffA;
    float *coeffB;
    float *x;
    float *y;
};

struct _channel_blk {
    struct _filter_blk  filter;
    struct _mixer_blk   mixer;        
} channel_blk;

//-----------------------------------------------------------------------------------
//  KEY/SIGNAL FUNCTIONS

sig_atomic_t ctrlC_pressed = 0;

void SIGINT_handler(int signum)
{
   assert (signum == SIGINT);
   ctrlC_pressed = 1;
}


//-----------------------------------------------------------------------------------
//  VSIB FUNCTIONS

int vsib_fileno = 0;
int vsib_mode = 0;                      // Mode of VSIB board, see modes.txt
int vsib_mode_gigabit = 0;              // 0: gigabit mode off
int vsib_mode_embed_1pps_markers = 0;   // 0: don't embed 1pps markers
int vsib_mode_skip_samples = 0;         // 0: don't skip samples, 1: leave out every 2nd sample, 2: ...

// vsib_ioctl(..) : A protected (error-checking) ioctl() for VSIB driver
static void vsib_ioctl(unsigned int mode, unsigned long arg)
{
  if (ioctl(vsib_fileno, mode, arg)) {
    char err[255];
    snprintf(err, sizeof(err), "rtspectrum: ioctl(vsib_fileno=%d, mode=0x%04x,...)", vsib_fileno, mode);
    perror(err);
    fprintf(stderr, "rtspectrum: standard I/O is not an VSIB board\n");
    exit(EXIT_FAILURE);
  }
}

//
// start_VSIB() : initialize VSIB using settings already in the global variables
//
void start_VSIB(void)
{
  vsib_ioctl(VSIB_SET_MODE,
             (VSIB_MODE_MODE(vsib_mode)
              | VSIB_MODE_RUN
              | (vsib_mode_gigabit ? VSIB_MODE_GIGABIT : 0)
              | (vsib_mode_embed_1pps_markers ? VSIB_MODE_EMBED_1PPS_MARKERS : 0)
              | (vsib_mode_skip_samples & 0x0000ffff))
            );
}

#define fourCharLong(a,b,c,d) ( ((long)a)<<24 | ((long)b)<<16 | ((long)c)<<8 | d )
typedef struct sSh {
  int relSeekBlocks;
} tSh, *ptSh;



//-----------------------------------------------------------------------------------
//  MAIN

int main(int argc, char **argv)
{
  int blksize;  /* memory / I/O block in traditional bytes */
  int blkwords;  /* mem / I/O in 32-bit words */
  int fftsize;
  int ffts_per_block;
  int totalblks;  /* how many blocks to read from file */
  char  *memblk;  /* file I/O pointer */
  char  *vsib_in;
  float *channel_out;
  int k;
  vsi_t prevMemwrd = 0;  /* previous VSI mem word for NRZM-decoding */
  size_t nread = 0, remaining = 0;
  struct sigaction sa;
  
  /* Shared memory unique identifier (key) and the returned reference id. */
  key_t shKey;
  int shId = -1;
  ptSh sh;

  /* Info / usage */
  printf("\nSBBC v1.0 -- a software baseband converter tool (DSP)\n");

  /* File read block size and the corresponding memblk of 32-bit words. */
  blkwords = 800;
  blksize = blkwords * sizeof(vsi_t);

  assert( (memblk = malloc(blksize)) != NULL );
  assert( (channel_out = malloc(blkwords * sizeof(float))) != NULL );

  #ifdef VSIB_DEV_FROMCLI
  vsib_fileno = STDIN_FILENO;
  #else
  vsib_fileno = open(VSIB_DEV_PATH, O_RDONLY);
  #endif

  #ifndef VSIB_DEV_FROMCLI
  /* disable echoing keyboard input */
  {
  static struct termios otty, ntty;
  tcgetattr(STDIN_FILENO, &otty);
  ntty = otty;
  ntty.c_lflag &= ~ECHO;
  tcsetattr(STDIN_FILENO, TCSANOW, &ntty);
  }
  #endif

  /* Register new SIGINT handler */
  sigemptyset (&sa.sa_mask);
  sa.sa_flags = 0;
  sa.sa_handler = SIGINT_handler;
  sigaction (SIGINT, &sa, 0);

  /* Create and initialize 'sbbc<-->control' shared memory. */
  shKey = fourCharLong('v','s','i','b');
  assert( (shId = shmget(shKey, sizeof(tSh), IPC_CREAT | 0777)) != -1 );
  assert( (sh = (ptSh)shmat(shId, NULL, 0)) != (void *)-1 );
  sh->relSeekBlocks = 0;

  /* Start VSIB board */
  vsib_mode = 0;  
  vsib_mode_gigabit = 0;
  vsib_mode_embed_1pps_markers = 0;
  vsib_mode_skip_samples = 0;
  start_VSIB();

  /* Init filter and mixer */
  
  
  while (!ctrlC_pressed) {
  
    /* read one full block */
    remaining = blksize;
    vsib_in   = memblk;
    do {
      nread = read (vsib_fileno, vsib_in, remaining);
      remaining -= nread;
      vsib_in += nread;
      if (remaining>0) {
         usleep(1000); // wait for buffer to fill; usleep resolution is 10ms?
      }
    } while ((remaining>0) && !ctrlC_pressed);

    /* apply filter */
    
    /* dump to STDOUT */  

  }

  /* Stop the board, first DMA, and when the last descriptor */
  /* has been transferred, then write stop to board command register. */
  vsib_ioctl(VSIB_DELAYED_STOP_DMA, 0);
  {
    unsigned long b = 0, usleeps = 0;

    vsib_ioctl(VSIB_IS_DMA_DONE, (unsigned long)&b);
    while (!b) {
      fprintf(stderr, "Waiting for last DMA descriptor (sl=%d)\n",
              usleeps);
      usleep(100000);
      usleeps++;
      vsib_ioctl(VSIB_IS_DMA_DONE, (unsigned long)&b);
    }
  }
  vsib_ioctl(VSIB_SET_MODE, VSIB_MODE_STOP);

  /* Remove shared memory to mark that 'wr/rd' is no more running. */
  if ((shId != -1) && (sh != (ptSh)-1) && (sh != NULL)) {
    assert( shmctl(shId, IPC_RMID, NULL) == 0 );
    assert( shmdt(sh) == 0 );
  }  // if shared memory was allocated

  #ifndef VSIB_DEV_FROMCLI
  close(vsib_fileno);
  #endif

  #ifndef VSIB_DEV_FROMCLI
  /* enable echoing keyboard input again */
  {
  static struct termios otty, ntty;
  tcgetattr(STDIN_FILENO, &otty);
  ntty = otty;
  ntty.c_lflag |= ECHO;
  tcsetattr(STDIN_FILENO, TCSANOW, &ntty);
  }
  #endif

  fprintf(stderr, "Done!\n");
  return 0;
}  /* main */



/*
 * $Log: sbbc.c,v $
 *
 */

