/*************************************************
 *                                               *
 * Simple Network UDP data to RAID receiver      *
 * (C) 2007 Jan Wagner                           *
 *                                               *
 * Listens on the specified port for UDP packets *
 * of user-specified size and writes them to a   *
 * specified output file.                        *
 * No packet filtering or user header inspection *
 * is done. In a sense, this is a dumb 'tcpdump' *
 * directly to disk.                             * 
 *                                               *
 * Usage: udp2raid port packetsize datasize file *
 *                                               *
 *   port : port of incoming UDP data            *
 *   packetsize : size of the UDP packets        *
 *   datasize : file size, how many bytes to get *
 *   file : output file                          *
 *                                               *
 *************************************************/

//
// Compile
//   gcc -Wall udp2raid.c -o udp2raid -D_LARGEFILE64_SOURCE=1
//
// Test
//   dd if=/dev/zero bs=32k count=40960 | pv -L 100m | netcat -u 193.166.42.22 9000
//    -- here 'pv' splits the data into random 1..N-byte sized writes
//       additionally netcat splits data into at most MTU sized UDP packets (e.g. max 8192 at 9000 MTU)
//

#define _GNU_SOURCE 1
#define _LARGEFILE64_SOURCE 1  /* Large File Support (LFS) '*64()' functions. */
#define _FILE_OFFSET_BITS 64   /* Automatic '*()' --> '*64()' replacement. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <fcntl.h>
#include <unistd.h>

#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <arpa/inet.h>
#include <netdb.h>
#include <sys/socket.h>

#define EXIT_OK          0
#define EXIT_FILE_ERROR  1
#define EXIT_NET_ERROR   2

#define MIN(a,b) (a<b) ? a : b
typedef unsigned long long ull_t;

double timedelta_us(struct timeval b, struct timeval e) {
   return ( 1e6*(e.tv_sec - b.tv_sec) - b.tv_usec + e.tv_usec );
}

int main(int argc, char* argv[]) {

  int      udpport;
  int      packetsize;
  int      smallest_intv_packetsize;
  int      smallest_rxd_packetsize;
  size_t   filesize;
  size_t   bytesremaining;
  size_t   packetcount;

  int      file_fd;
  int      socket_fd;
  char*    mmbuf;
  char*    mmptr;

  struct addrinfo  hints;
  struct addrinfo *info;
  struct addrinfo *info_save;
  char             aibuffer[10];
  int              status;

  struct timeval   tv_xferstart, tv_xferstop;
  struct timeval   tv_intvstart, tv_intvstop;
  size_t           intvbytes;
  double           tvdiff;

  if (argc < 5) {
     printf("Usage: udp2raid port packetsize datasize file\n\n"
            "   port : port of incoming UDP data\n"
            "   packetsize : size in bytes of the UDP packet read()\n"
            "   datasize : how many bytes to collect in total\n"
            "   file : name and path of the output file\n\n");
     return EXIT_OK;
  }


  /* get user params */
  udpport    = atoi(argv[1]);
  packetsize = atoi(argv[2]);
  filesize   = atol(argv[3]);  


  /* open the file */
  file_fd = open(argv[4], O_RDWR|O_CREAT/*|O_TRUNC*/, 
                 S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH);
  if (-1 == file_fd) {
     fprintf(stderr, "File open failed: '%s'\n", strerror(errno)); 
     return EXIT_FILE_ERROR;
  }  


  /* stretch the file to final length */
  fprintf(stderr, "File opened, stretching to %Lu bytes...\n", (ull_t)filesize);
  if (-1 == lseek(file_fd, filesize-1, SEEK_SET)) {
     fprintf(stderr, "File streching with lseek failed: '%s'\n", strerror(errno)); 
     return EXIT_FILE_ERROR;
  }
  write(file_fd, "empty", 1);


  /* map the file */
  mmbuf = (char*)mmap((caddr_t)0, filesize, PROT_WRITE|PROT_READ, MAP_SHARED, file_fd, 0);
  if (MAP_FAILED == mmbuf) {
     fprintf(stderr, "File mmap failed: '%s'\n", strerror(errno)); 
     return EXIT_FILE_ERROR;
  }


  /* bind UDP socket */
  memset(&hints, 0, sizeof(hints));
  hints.ai_flags    = AI_PASSIVE;
  hints.ai_family   = AF_INET;
  hints.ai_socktype = SOCK_DGRAM;

  sprintf(aibuffer, "%u", udpport);
  getaddrinfo(NULL, aibuffer, &hints, &info);

  info_save = info; status = 0;
  do {
     socket_fd = socket(info->ai_family, info->ai_socktype, info->ai_protocol);
     if (socket_fd < 0) { continue; }
     if (!bind(socket_fd, info->ai_addr, info->ai_addrlen)) {
         fprintf(stderr, "Receiving UDP on port %u\n", udpport);
         status = 1;
         break;
     }
  } while ((info = info->ai_next) != NULL);
  freeaddrinfo(info_save);
  
  if (0 == status) {
     fprintf(stderr, "Couldn't get a socket for port %u\n", udpport);
     return EXIT_NET_ERROR;
  }


  /* prepare vars */
  mmptr                    = mmbuf; 
  intvbytes                = bytesremaining;
  packetcount              = 0;
  bytesremaining           = filesize;
  smallest_rxd_packetsize  = packetsize;
  smallest_intv_packetsize = smallest_rxd_packetsize;

  /* receive all the data into the file/memory */
  while (bytesremaining > 0) {
     ssize_t rc, rdlen;

     // receive into the file, crop at file end
     rdlen = MIN(packetsize, bytesremaining);
     rc    = recv(socket_fd, mmptr, rdlen, 0);
     if (rc < 0) {
        fprintf(stderr, "recv() at byte %p failed: '%s'\n", 
                (void*)(mmptr-mmbuf), strerror(errno));
        break;
     }

     // remember smallest packet
     if ((rdlen == packetsize) && (rc < smallest_intv_packetsize)) {
        smallest_intv_packetsize = rc;
     }

     // move forward
     bytesremaining -= rc;
     mmptr += rc;

     // start timing?
     if (0 == packetcount) {
        fprintf(stderr, "   Bytes left | Rate Mbit/s | Smallest UDP\n");
        gettimeofday(&tv_xferstart, NULL);
        tv_intvstart = tv_xferstart;
     }
     packetcount++;

     // interval reports
     if (0 == (packetcount & 0x3FFF)) {

        gettimeofday(&tv_intvstop, NULL);
        tvdiff = timedelta_us(tv_intvstart, tv_intvstop);

        fprintf(stderr, " %12Lu | %11.3f | %6u\n", 
                (ull_t)bytesremaining, 
                8.0e6*(intvbytes-bytesremaining)/(1024.0*1024.0*tvdiff),
                smallest_intv_packetsize );

        if (smallest_intv_packetsize < smallest_rxd_packetsize) {
            smallest_rxd_packetsize = smallest_intv_packetsize;
        }

        gettimeofday(&tv_intvstart, NULL);
        intvbytes = bytesremaining;
        smallest_intv_packetsize = packetsize;
     }
  }


  /* close the socket */
  close(socket_fd);

  /* release with intrinsic flush */
  munmap((caddr_t)mmbuf, filesize);

  /* final stats */
  gettimeofday(&tv_xferstop, NULL);
  tvdiff = timedelta_us(tv_xferstart, tv_xferstop);
  fprintf(stderr, "Completed. Received %Lu packets, the smallest was %u bytes long.\n",
          (ull_t)packetcount, smallest_rxd_packetsize);
  fprintf(stderr, "%8.3f MB in %8.2f us : %8.3f Mbit/s\n",
          filesize/(1024.0*1024.0), tvdiff, 
          8.0e6*filesize/(1024.0*1024.0*tvdiff));

  return EXIT_OK;
}
