/* wr.c -- Read/write VLBI VSIB input/output board and disks. */

/*
 * $Log: wr.c,v $
 * Revision 1.17  2006/10/04 07:07:27  jwagner
 * increased stdio pipe buffer to 2MB
 *
 * Revision 1.16  2006/09/29 12:08:36  jwagner
 * file name - for stdout or stdin instead of disk
 *
 * Revision 1.15  2006/02/23 09:50:40  amn
 * Adding basic pre-release software documents.
 *
 * Revision 1.14  2004/08/02 18:35:39  amn
 * First release of sf.c, software Mk4/VLBA formatter.
 *
 * Revision 1.13  2003/03/05 08:01:56  amn
 * March 2003 Jb2/Mh Mk4/5-style Gbit test version.
 *
 * Revision 1.12  2002/11/29 10:58:40  amn
 * Unified ft2 and ft3; made mrg seq. numbered file opening work; wr parms.
 *
 * Revision 1.11  2002/11/07 09:06:49  amn
 * First Mk5A 1Gbit/s tests with Jodrell.
 *
 * Revision 1.10  2002/11/01 13:19:58  amn
 * Just before changing to use a common tstamp.cc file.
 *
 * Revision 1.9  2002/09/20 07:39:15  amn
 * Added mrg.c, merge/split for multi-PC files.
 *
 * Revision 1.8  2002/09/02 16:39:47  amn
 * Test version suitable for 50MHz/2 test pattern recording, for pb/JB.
 *
 * Revision 1.7  2002/08/09 11:26:56  amn
 * Jul-2002 first fringes Dwingeloo test version.
 *
 * Revision 1.5  2002/06/15 10:07:31  amn
 * Added shared memory and seek support.
 *
 * Revision 1.4  2002/06/14 13:02:34  amn
 * Dwingeloo test trip version.
 *
 */

/* Copyright (C) 2001--2002 Ari Mujunen, Ari.Mujunen@hut.fi

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */


/* Ensure 64-bit file support regardless of '-D_LARGEFILE64_SOURCE=1'
   in Makefile. */
#define _LARGEFILE64_SOURCE 1  /* Large File Support (LFS) '*64()' functions. */ 
#define _FILE_OFFSET_BITS 64  /* Automatic '*()' --> '*64()' replacement. */ 

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include <sys/types.h>  /* open() */
#include <sys/stat.h>  /* open() */
#include <fcntl.h>  /* open() */

#include <sys/time.h>  /* gettimeofday() */
#include <unistd.h>  /* gettimeofday(), usleep() */

#include <string.h>  /* strstr() */

#include <sys/ipc.h>  /* for shared memory */
#include <sys/shm.h>

#include <sys/ioctl.h>
#include "vsib_ioctl.h"

/* Depending on readMode either standard input or output. */
int vsib_fileno;

/* A protected (error-checking) ioctl() for VSIB driver. */
static void
vsib_ioctl(
  unsigned int mode,
  unsigned long arg
) {
  if (ioctl(vsib_fileno,
            mode,
            arg)
  ) {
    char *which;
    char err[255];
    
    if (vsib_fileno == STDOUT_FILENO) {
      which = "rd";
    } else {
      which = "wr";
    }
    snprintf(err, sizeof(err), "%s: ioctl(vsib_fileno, 0x%04x,...)", which, mode);
    perror(err);
    fprintf(stderr, "%s: standard I/O is not an VSIB board\n", which);
    exit(EXIT_FAILURE);
  }
}  /* vsib_ioctl */


typedef struct sDir {
  char *nameTemplate;
  int nameNumber;
  int blocksPerFile;
  int fid;
  char useDisk;  
  int nowBlocks;
  double prevOperTime;
  double maxOperTime;
} tDir, *pDir;


double
tim(void) {
  struct timeval tv;
  double t;

  assert( gettimeofday(&tv, NULL) == 0 );
  t = (double)tv.tv_sec + (double)tv.tv_usec/1000000.0;
  return t;
}  /* tim */


void
randbuf(char *buf, ssize_t nbytes)
{
  int i;

  for (i=0; i<nbytes; i++) {
    buf[i] = (char)rand();
  }  /* for */
}  /* randbuf */


int vsib_mode;
int vsib_mode_gigabit;
int vsib_mode_embed_1pps_markers;
int vsib_mode_skip_samples;

void
start_VSIB(void)
{
  vsib_ioctl(VSIB_SET_MODE,
             (VSIB_MODE_MODE(vsib_mode)
              | VSIB_MODE_RUN
              | (vsib_mode_gigabit ? VSIB_MODE_GIGABIT : 0)
              | (vsib_mode_embed_1pps_markers ? VSIB_MODE_EMBED_1PPS_MARKERS : 0)
              | (vsib_mode_skip_samples & 0x0000ffff))
            );
}  /* start_VSIB */


/* Upcoming wrshm.h: */
#define fourCharLong(a,b,c,d) ( ((long)a)<<24 | ((long)b)<<16 | ((long)c)<<8 | d )
typedef struct sSh {
  int relSeekBlocks;
} tSh, *ptSh;



int
main(int argc, char *argv[]) {
  int readMode;
  int blksize;
  int totalblks;
  int memblks;
  char *memblk;
  char *p;
  int ndirs;
  pDir dirs;
  int i;
  double recordStart;
  double totalDur;
  int usleeps;
  int vsib_started = 0;
  int blk;
  double mbits;

  /* Shared memory unique identifier (key) and the returned reference id. */
  key_t shKey;
  int shId = -1;
  ptSh sh;

 char *stdIO_buffer;

  if (argc < 8) {
    fprintf(stderr, "%s: needs at least blocksize totalblocks mode skip embed giga ndirs { path%%d blks }\n", argv[0]);
    exit(EXIT_FAILURE);
  }

#define TIMESTART(i) \
      { dirs[i].prevOperTime = tim(); }
#define TIMESTOP(i) \
      { double dur = tim() - dirs[i].prevOperTime; \
        if (dur > dirs[i].maxOperTime) dirs[i].maxOperTime = dur; }

  /* Init variables according to command line, allocate buffers. */
  readMode = (strstr(argv[0], "rd") != NULL);
  if (readMode) {
    /* Read from disk, write to stdout to VSIB board. */
    vsib_fileno = STDOUT_FILENO;
  } else {
    vsib_fileno = STDIN_FILENO;
  }
  blksize = atoi(argv[1]);
  totalblks = atoi(argv[2]);
  /* memblks = atoi(argv[ ]); */
  memblks = 1;
  assert( (memblk = malloc(memblks*blksize)) != NULL );
#if 0
  /* Init memblk with random values. */
  {
    double st = tim();

    fprintf(stderr, "Initializing memory buffer with rand()...");
    fflush(stdout);
    randbuf(memblk, (memblks*blksize));
    fprintf(stderr, "took %f seconds.\n", tim()-st);
  }
#endif
  vsib_mode = atoi(argv[3]);
  vsib_mode_skip_samples = atoi(argv[4]);
  vsib_mode_embed_1pps_markers = atoi(argv[5]);
  vsib_mode_gigabit = atoi(argv[6]);
  ndirs = atoi(argv[7]);
  assert( (argc >= (8 + ndirs*2)) );
  /* Allocate one extra struct for call-to-call time statistics. */
  assert( (dirs = malloc((ndirs+1)*sizeof(tDir))) != NULL );
  for (i=0; i<ndirs; i++) {
    dirs[i].nameTemplate = argv[8+i*2];
    dirs[i].nameNumber = 0;
    dirs[i].blocksPerFile = atoi(argv[9+i*2]);
    dirs[i].nowBlocks = dirs[i].blocksPerFile;  /* init to end */
    dirs[i].maxOperTime = 0.0;  /* init to smallest value */
    dirs[i].useDisk = (0!=strcmp(argv[8+i*2], "-"));
  }  /* for each directory */
  dirs[ndirs].nameTemplate =   "call-to-call  ";
  dirs[ndirs].maxOperTime = 0.0;

  /* Create and initialize 'wr<-->control' shared memory. */
  shKey = fourCharLong('v','s','i','b');
  assert( (shId = shmget(shKey, sizeof(tSh), IPC_CREAT | 0777)) != -1 );
  assert( (sh = (ptSh)shmat(shId, NULL, 0)) != (void *)-1 );
  sh->relSeekBlocks = 0;

#if 0
  Opening the device special file already does this.
  For write() this open()-related (extra) DMA reset clears
  memory buffers and r/w pointers with arg==0.
  /* Init DMA. */
  vsib_ioctl(VSIB_RESET_DMA, 0);
#endif

  /* Start the board.  (After opening it is still stopped.) */
  /* For disk write, start VSIB first and write to disk later. */
  /* For disk read, read first and start VSIB later. */
  if (readMode) {
    vsib_started = 0;
  } else {
    start_VSIB();
  }

  /* Start looping for total number of blocks. */
  usleeps = 0;
  recordStart = tim();
  TIMESTART(ndirs);  /* call-to-call (not per file) */
  for (blk=0; blk<totalblks; blk++) {

    i = blk % ndirs;
    if (sh->relSeekBlocks) {
      int bl;

      /* Adjust 'blk' (and i and newBlocks) based on 'sh->relSeekBlocks'. */
      /* xxx: initially only within a single open file */
      bl = dirs[i].nowBlocks + sh->relSeekBlocks;
      if (bl < 0) {
	sh->relSeekBlocks -= bl;
	bl = 0;
      }
      bl = dirs[i].nowBlocks + sh->relSeekBlocks;
      if (bl > dirs[i].blocksPerFile) {
	sh->relSeekBlocks -= (bl - dirs[i].blocksPerFile);
	bl = dirs[i].blocksPerFile;
      }
      assert( lseek(dirs[i].fid, (bl*blksize), SEEK_SET) != (off_t)-1 );
      fprintf(stderr, "at block = %d, repositioning %+d blocks (to byte offset %d)\n",
	      blk,
	      sh->relSeekBlocks,
	      (bl*blksize)
	      );
      dirs[i].nowBlocks = bl;
      blk += sh->relSeekBlocks;
      /* xxx: cannot really update i... */

      /* Seeking done, notify shared memory. */
      sh->relSeekBlocks = 0;
    }  /* if external adjustments */

    if (dirs[i].nowBlocks >= dirs[i].blocksPerFile) {
      /* Open a new file in this directory / template. */
      char pn[255];
      int n = blk / dirs[i].blocksPerFile;

      snprintf(pn, sizeof(pn), dirs[i].nameTemplate, n);
      if (n==0) TIMESTART(i);  /* only for first file */
      if(dirs[i].useDisk) {
         if (readMode) {
           assert( (dirs[i].fid = open(pn, O_RDONLY)) != -1 );
         } else {
           assert( (dirs[i].fid = creat(pn, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) != -1 );
      /* O_SYNC here doesn't change 2.4.xx buffer flush behavior. */
         }
         fprintf(stderr, "at block = %d, opened file '%s'\n", blk, pn);
      } else {
         dirs[i].fid = readMode ? STDIN_FILENO : STDOUT_FILENO;
         #define STDIO_NEW_BUFLEN   2*1024*1024
         stdIO_buffer = malloc(STDIO_NEW_BUFLEN);
         if (readMode) {
            setvbuf(stdin, stdIO_buffer, _IOFBF,  STDIO_NEW_BUFLEN); // _IOFBF = fully buffer the I/O
         } else {
            setvbuf(stdout, stdIO_buffer, _IOFBF,  STDIO_NEW_BUFLEN);
         }
         fprintf(stderr, "at block = %d, concatenating next over stdio (fid=%d)'\n", blk, dirs[i].fid);
      }
      dirs[i].nowBlocks = 0;
    }

    /* Write (or read) one block. */
    /* xxx: need to add '-1' error checks to VSIB read()/write() calls */
    p = &(memblk[(blk % memblks)*blksize]);
    if (readMode) {
      size_t nwritten;

      assert( (read(dirs[i].fid,
                     p,
                     blksize)) == blksize );

      /* Write a block to VSIB; if not enough space, sleep a little. */
      nwritten = write (vsib_fileno, memblk, blksize);
      while (nwritten < blksize) {
        /* Not space for one full block in buffer, start VSIB transmission */
        /* and wait for vacant space. */
        if (readMode) {
          if (!vsib_started) {
            fprintf(stderr, "starting vsib because bigbuf full\n");
            /* xxx: reload PLX FIFOs */
            vsib_ioctl(VSIB_RESET_DMA, /*skip_clear=>*/1);
            start_VSIB();
            vsib_started = 1;
          }
        }
        usleep(1000);  /* a small amount, probably ends up to be 10--20msec */
        usleeps++;
        nwritten += write (vsib_fileno, memblk+nwritten, blksize-nwritten);
      }  /* while not at least one full block in VSIB DMA ring buffer */

    } else {
      size_t nread;

      /* Read a block from VSIB; if not enough, sleep a little. */
      nread = read (vsib_fileno, memblk, blksize);
      while (nread < blksize) {
        /* Not one full block in buffer, wait. */
        usleep(1000);  /* a small amount, probably ends up to be 10--20msec */
        usleeps++;
        nread += read (vsib_fileno, memblk+nread, blksize-nread);
      }  /* while not at least one full block in VSIB DMA ring buffer */

      assert( (write(dirs[i].fid,
                     p,
                     blksize)) == blksize );
    }
    TIMESTOP(i);  /* accumulates close--open-write in worst case */
    TIMESTOP(ndirs);  /* call-to-call */

    TIMESTART(i);
    TIMESTART(ndirs);
    dirs[i].nowBlocks++;

    /* Time to switch to another file? */
    if (dirs[i].nowBlocks >= dirs[i].blocksPerFile) {
      /* Close the ready file before opening a new. */
      if (dirs[i].useDisk) {
         assert( (close(dirs[i].fid) != -1) );
      }
    }
  }  /* for each block */
  TIMESTOP(ndirs);
  totalDur = tim() - recordStart;

  /* Wait for the data to get out of VSIB. */
  if (readMode) {
    unsigned long prevb;
    unsigned long b;

    if (!vsib_started) {
      fprintf(stderr, "all written to bigbuf, starting vsib...\n");
      /* xxx: reload PLX FIFOs */
      vsib_ioctl(VSIB_RESET_DMA, /*skip_clear=>*/1);
      start_VSIB();
      /* usleep(2000000); */  /* xxx */
    }
#if 1
    vsib_ioctl(VSIB_GET_BYTES_IN_BIGBUF, (unsigned long)&b);
    prevb = b;
    while (b <= prevb) {
      fprintf(stderr, "Waiting for ring buffer to empty (%lu bytes, sl=%d)\n",
              b, usleeps);
      /* Still bytes in bigbuf waiting for transmit. */
      usleep(10000);  /* 0.01sec */
      usleeps++;
      prevb = b;
      vsib_ioctl(VSIB_GET_BYTES_IN_BIGBUF, (unsigned long)&b);
    }  /* while not bigbuf empty */
#endif    
  }  /* if readMode, wait for data to get out of ring buffer */

  /* Stop the board, first DMA, and when the last descriptor */
  /* has been transferred, then write stop to board command register. */
  vsib_ioctl(VSIB_DELAYED_STOP_DMA, 0);
  {
    unsigned long b;

    vsib_ioctl(VSIB_IS_DMA_DONE, (unsigned long)&b);
    while (!b) {
      fprintf(stderr, "Waiting for last DMA descriptor (sl=%d)\n",
              usleeps);
      usleep(100000);
      usleeps++;
      vsib_ioctl(VSIB_IS_DMA_DONE, (unsigned long)&b);
    }
  }
  vsib_ioctl(VSIB_SET_MODE, VSIB_MODE_STOP);

  /* Final report. */
  mbits = 8.0*totalblks*blksize/totalDur/1000.0/1000.0;
  fprintf(stderr, "Took %f seconds, %f Mbits(dec)/s (%.1f%% of PCI33), usleeps = %d.\n",
         totalDur,
         mbits,
         mbits/(33.0*4.0*8.0)*100.0,
         usleeps
         );
  for (i=0; i<=ndirs; i++) {
    fprintf(stderr, "%s  max. %f seconds.\n",
           dirs[i].nameTemplate,
           dirs[i].maxOperTime
           );
  }  /* for each directory */

  /* Remove shared memory to mark that 'wr/rd' is no more running. */
  if ((shId != -1) && (sh != (ptSh)-1) && (sh != NULL)) {
    assert( shmctl(shId, IPC_RMID, NULL) == 0 );
    assert( shmdt(sh) == 0 );
  }  // if shared memory was allocated

  return(EXIT_SUCCESS);
}  /* main */
