/*------------------------------------------------------
 * 
 *  wrimp.c 
 *
 *  An improved 'wr' for writing scans to disk or stdout,
 *  Lacks some of the features of 'wr' but adds others.
 *
 *------------------------------------------------------
 * $Log: wrimp.c,v $
 * Revision 1.1  2007/10/08 08:56:54  jwagner
 * improved but untested new wr
 *
 *
 */

/* Copyright (C) 2001--2002 Ari Mujunen, Ari.Mujunen@hut.fi
   Copyright (C) 2007--2008 Jan Wagner, jwagner@kurp.hut.fi

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */


/* Ensure 64-bit file support regardless of '-D_LARGEFILE64_SOURCE=1' in Makefile. */
#define _LARGEFILE64_SOURCE 1  /* Large File Support (LFS) '*64()' functions. */ 
#define _FILE_OFFSET_BITS 64  /* Automatic '*()' --> '*64()' replacement. */ 

#include <stdio.h>
#include <stdlib.h>     /* memalign() etc */
#include <assert.h>
#include <signal.h>     /* Control-C int reassign */

#include <sys/types.h>  /* open() */
#include <sys/stat.h>   /* open() */
#include <fcntl.h>      /* open() */

#include <sys/time.h>   /* gettimeofday() */
#include <unistd.h>     /* gettimeofday(), usleep(), getopt() */
#include <getopt.h>     /* getopt_long() */

#include <string.h>     /* strstr() */

#include <sys/ipc.h>    /* for shared memory */
#include <sys/shm.h>

#include <sys/ioctl.h>

#include "common/parse_evn_filename.h"
#include "common/parse_evn_filename.c"
#include "vsib_ioctl.h"

extern char *optarg;
extern int optind, opterr, optopt;

//=========================================================================================================
// GLOBALS AND DECLARES
//=========================================================================================================

#define READ_BUF_SIZE   131072
#define VSIB_DEV_PATH    "/dev/vsib"  // hard-coded reference to VSIB device node
int vsib_fileno;
sig_atomic_t ctrlC_pressed = 0;

//=========================================================================================================
// HELPERS
//=========================================================================================================

/* Catch Control-C */
void SIGINT_handler(int signum)
{
   assert (signum == SIGINT);
   ctrlC_pressed = 1;
}

/* A protected (error-checking) ioctl() for VSIB driver. */
void vsib_ioctl(unsigned int mode, unsigned long arg) 
{
  if (ioctl(vsib_fileno, mode, arg)) {
    char err[255];
    snprintf(err, sizeof(err), "ioctl(vsib_fileno, 0x%04x,...)", mode);
    perror(err);
    fprintf(stderr, "File is is not an VSIB board\n");
    exit(EXIT_FAILURE);
  }
}

void start_VSIB(int mode, int giga, int embed, int skip)
{
  vsib_ioctl(VSIB_SET_MODE,
             (VSIB_MODE_MODE(mode)
              | VSIB_MODE_RUN
              | (giga  ? VSIB_MODE_GIGABIT : 0)
              | (embed ? VSIB_MODE_EMBED_1PPS_MARKERS : 0)
              | (skip & 0x0000ffff))
            );
}

void stop_VSIB(void)
{
    vsib_ioctl(VSIB_SET_MODE, VSIB_MODE_STOP);
}

double dsecs(void) {
  struct timeval tv;
  assert( gettimeofday(&tv, NULL) == 0 );
  /* 64-bit double rounds off least significant microseconds
   * if seconds difference is large. */
  return ((double)tv.tv_sec + (double)tv.tv_usec/1000000.0);
}

//=========================================================================================================
// MAIN
//=========================================================================================================

int main(int argc, char *argv[]) {
  int   fvsib;
  int   fdisk;
  char* buf;
  char* evnfilename;
  char *duration;
  struct evn_filename * pevnf;
  struct sigaction sa;

  int usleeps = 0;
  double t_start, t_stop;
  double now, starttime;

  int vsib_mode  = 0;
  int vsib_giga  = 0;
  int vsib_embed = 0;
  int vsib_skip  = 0;
  int vsib_mbps  = 512;
  int verbose = 0;
  int skip_old = 0;

  float seconds;
  double skip_adjust;
  unsigned long long total_bytes, bytes_left;
  unsigned long vsib_fillstate_curr, vsib_fillstate_peak = 0;

  struct option long_options[] = { 
        { "verbose",  0, NULL, 1 },
        { "rate",     1, NULL, 2 }, /* Mbit/s */
        { "mode",     1, NULL, 3 }, /* VSIB mode */
        { "skip",     1, NULL, 4 }, /* Read 1 skip N */
        { "skipold",  0, NULL, 5 }, /* Skip recording is start time is in the past */
        { NULL,       0, NULL, 0 } 
  };
  int opt;

  while ((opt = getopt_long(argc, argv, "+", long_options, NULL)) > 0) {
    switch(opt) {
        case 1:
            verbose = 1;
            break;
        case 2:
            vsib_mbps = atoi(optarg);
            if (!vsib_mbps) exit(printf("Rate given (%s) is zero\n", optarg));
            break;
        case 3:
            vsib_mode = atoi(optarg);
            break;
        case 4:
            vsib_skip = atoi(optarg);
            break;
        case 5:
            skip_old = 1;
            break;
    }
  }

  if ((int)(argc-optind) < 2) {
     printf(" wrimp [--verbose] [--rate=512] [--mode=x] [--skip=0] [--skipold] EVNfilename seconds\n\n");
     exit(EXIT_FAILURE);
  }

  evnfilename = argv[optind];
  duration = argv[optind+1];
  sscanf(duration, "%f", &seconds);

  skip_adjust = 1.0 / (1.0 + (double)vsib_skip);
  total_bytes = (double)seconds * (double)(1024*1024) * (double)(vsib_mbps/8) * skip_adjust;
  bytes_left  = total_bytes;

  if (verbose) {
    fprintf(stderr, "--VSIB-------------------------------------\n");
    fprintf(stderr, "Mode : %d\n", vsib_mode);
    fprintf(stderr, "Skip : %d samples (adjust factor = %f)\n", vsib_skip, skip_adjust);
    fprintf(stderr, "--SCAN-------------------------------------\n");
    fprintf(stderr, "Rate : %d Mbit/s\n", vsib_mbps);
    fprintf(stderr, "File : write to '%s'\n", evnfilename);
    fprintf(stderr, "Past : %s", (skip_old) ? "skip passed scans\n" : "still record passed scans\n");
    fprintf(stderr, "Time : record %f seconds\n", seconds);
    fprintf(stderr, "--CALCULATED-------------------------------\n");
    fprintf(stderr, "Read : %lld bytes\n", total_bytes);
    fprintf(stderr, "-------------------------------------------\n");
  }


  /* Register new SIGINT handler to catch Control-C */
  sigemptyset (&sa.sa_mask);
  sa.sa_flags = 0;
  sa.sa_handler = SIGINT_handler;
//  sigaction (SIGINT, &sa, 0);

  /* Allocate user read buffer */
  buf = (char*) memalign(16, READ_BUF_SIZE);
  if (NULL == buf) {  
     fprintf(stderr, "Error: Could not allocate buffer (%d bytes).\n", READ_BUF_SIZE);
     exit(EXIT_FAILURE);
  }

  /* Open the files */
  fvsib = open(VSIB_DEV_PATH, O_RDONLY);
  if (-1 == fvsib) {  
     fprintf(stderr, "Error: Could not open /dev/vsib!\n");
     exit(EXIT_FAILURE);
  }
  vsib_fileno = (int)fvsib;
  fdisk = creat(evnfilename, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
  if (-1 == fdisk) {
     fprintf(stderr, "Error: Could not create output file!\n");
     exit(EXIT_FAILURE);
  }

  /* Get and convert target UTC time to Unix seconds. */
  pevnf = parse_evn_filename(evnfilename);
  if ((0 == pevnf->valid) || (NULL == pevnf->data_start_time_ascii)) {
     fprintf(stderr, "Error: Filename doesn't follow the EVN Filenaming Convention.\n");
     exit(EXIT_FAILURE);
  }

  /* Wait until the start time */
  starttime = pevnf->data_start_time - 0.5;
  now = dsecs();
  if (now > starttime) {
     if (skip_old) exit(EXIT_FAILURE);
  }
  while (now < starttime) {
     double stillLeft = starttime - now;
     double newSleep;
     unsigned long sleepusecs;
     newSleep = stillLeft / 2;
     sleepusecs = (unsigned long)(newSleep * 1000000.0);
     fprintf(stderr, "%.2f seconds until `%s', sleeping %.2f seconds...\n", stillLeft, argv[1], newSleep);
     usleep(sleepusecs); /* accurate to at most 10ms, but that's enough */
     if (ctrlC_pressed) { 
        fprintf(stderr, "Cancelled with Control-C\n");
        exit(EXIT_FAILURE);
     }
     now = dsecs();
  }

  /* Start data acquisition */
  fprintf(stderr, "Starting to record.\n");
  start_VSIB(vsib_mode, vsib_giga, vsib_embed, vsib_skip);
  t_start = dsecs();
  while ((bytes_left > 0) && !ctrlC_pressed) {
      /* read vsib and write out*/
      size_t nread;
      size_t toread;
      if (bytes_left < READ_BUF_SIZE) {
        toread = bytes_left;
      } else {
        toread = READ_BUF_SIZE;
      }
      nread = read(fvsib, buf, toread);
      while (nread < toread) {
        usleeps++; usleep(1000); /* 10..20ms */
        nread += read(fvsib, buf+nread, toread-nread);
      }
      write(fdisk, buf, toread); // == toread;
      total_bytes -= toread; 
      /* monitor vsib buffer occupancy growth */
      vsib_ioctl(VSIB_GET_BYTES_IN_BIGBUF, (unsigned long)&vsib_fillstate_curr);
      if (vsib_fillstate_curr > vsib_fillstate_peak) {
        vsib_fillstate_peak = vsib_fillstate_curr;
        fprintf(stderr, "Peak available bytes grew to %lu bytes\n", vsib_fillstate_peak);
      }
  }
  t_stop = dsecs();
  stop_VSIB();

  close(fvsib);
  close(fdisk);
  if (ctrlC_pressed) 
      fprintf(stderr, "Cancelled with Control-C\n");

  fprintf(stderr, "Took %f seconds, %f Mbits(dec)/s \n",
         (t_stop-t_start), total_bytes / (t_stop-t_start) );

  return 0;
}
